#!/usr/bin/env python

from hh_reader import read_result
from copy import deepcopy
from pdbx.reader.PdbxReader import PdbxReader
from pdbx.writer.PdbxWriter import PdbxWriter
import re, os, sys, tempfile, glob

from operator import itemgetter  # hzhu
from itertools import groupby    # hzhu

EMPTY = '*'
GAP = '-'
DEBUG_MODE = False

class Gap: 
    """ A gap is a continuous stretch of indels.
        It is defined by a opening position and a size/length
    """
    def __init__(self, open_pos, size):
        self.open_pos = open_pos   # gap opening position
        self.size = size           # num of indels in the gap

    def __repr__(self):
        return 'Gap opening pos = %d, size = %d' % (self.open_pos, self.size)
        
class Grid:
    """
    Implementation of 2D grid of cells
    Includes boundary handling
    """
    
    def __init__(self, grid_height, grid_width):
        """
        Initializes grid to be empty, take height and width of grid as parameters
        Indexed by rows (left to right), then by columns (top to bottom)
        """
        
        self._grid_height = grid_height
        self._grid_width = grid_width
        self._cells = [ [ EMPTY for dummy_col in range(self._grid_width) ]
                       for dummy_row in range(self._grid_height)]
                
    def __str__(self):
        """ Return multi-line string represenation for grid """
        
        ans = ''
        for row in range(self._grid_height):
            ans += ''.join(self._cells[row])
            ans += '\n'
        return ans

    def clear(self):
        """ Clears grid to be empty """
        
        self._cells = [[EMPTY for dummy_col in range(self._grid_width)]
                       for dummy_row in range(self._grid_height)]
    
    def get_grid_height(self):
        """ Return the height of the grid """
        
        return self._grid_height

    def get_grid_width(self):
        """ Return the width of the grid """
        
        return self._grid_width
    
    def get_cell(self, row, col):
        return self._cells[row][col]

    def get_seq_start(self, row):
        """ Returns the start position of the sequence """
        
        index = 0
        for pos in self._cells[row]:
            if pos != EMPTY: 
                return index
            index += 1

        return None

    def get_seq_end(self, row):
        """ Returns the end position of the sequence """
        
        index = 0
        for pos in reversed(self._cells[row]):
            if pos != EMPTY: 
                return self.get_grid_width() - index
            index += 1

        return None

    def get_gaps(self, row): 
        """ Return the position of gaps in a row """
        
        gaps = list()

        index = 0
        for pos in self._cells[row]:
            if pos == GAP: 
                gaps.append(index)
            index += 1

        return gaps

    def get_gaps_ref_gapless(self, row):
        """ Return the pos of gaps in a row.
            The opening positions of the gaps are wrt. the gapless seq
        """
        # get all the indels
        indels = self.get_gaps(row)  
        gaps = []
        # combine continuous indels into a gap
        for k,i in groupby( enumerate(indels), lambda x: x[0]-x[1] ):
            g = list(map(itemgetter(1), i))
            gaps.append( Gap(g[0], len(g)) )

        # offset the gap opening positions
        for i in range(1, len(gaps)):
            # offset by total gap number before
            gaps[i].open_pos -= sum([gaps[j].size for j in range(i)])
            
        return gaps  # a list of Gap instances
    
    def get_seq_indeces(self, row):

        seq = list()
        for pos, res in enumerate(self._cells[row]):
            if res != EMPTY and res != GAP:
                seq.append(pos)

        return seq
            
    ## def get_gap_list(self):  # hzhu commented this out. wrote a new version
    ##     """ Returns a list of list of all gap positions in the sequence grid. """
    ##     gap_pos = set()

    ##     for row in range(self.get_grid_height()):
    ##         for gap in self.get_gaps(row):
    ##             gap_pos.add(gap)

    ##     gap_pos = list(sorted(gap_pos))
    ##     boundaries = [ (x + 1) for x, y in zip(gap_pos, gap_pos[1:]) if y - x != 1 ]
 
    ##     gap_list = list()
    ##     prev = 0
 
    ##     for boundary in boundaries:
    ##         sub_list = [ pos for pos in gap_pos[prev:] if pos < boundary ]
    ##         gap_list.append(sub_list)
    ##         prev += len(sub_list)
 
    ##     gap_list.append([ x for x in gap_pos[prev:]])
    
    ##     return gap_list

    def get_gap_list(self):
        """ Returns a list of Gap instances for all rows in the grid
        """
        gap_dict = dict()  # each position should occur as gap at most once
                           # keys are gap openning positions
                           # values are Gap instances
        gap_list = []
        for row in range(self.get_grid_height()):
            gap_pos = []
            gaps = self.get_gaps_ref_gapless(row)
             
            for g in gaps:
                if g.open_pos in gap_dict: # if there is already gaps at this open pos
                    if g.size > gap_dict[g.open_pos].size: # if new gap is bigger
                        gap_dict[g.open_pos] = g  # keep the larger gap as they overlap
                else:
                    gap_dict[g.open_pos] = g
                    
        gap_list = sorted(list(gap_dict.values()), key=lambda x: x.open_pos) # sort according to start position
        return gap_list  # a list of Gap instances
        
    def set_gap(self, row, col):
        """ Set cell with index (row, col) to be a gap """
        
        self._cells[row][col] = GAP

    def set_empty(self, row, col):
        """ Set cell with index (row, col) to be a gap """
        
        self._cells[row][col] = EMPTY
    
    def set_cell(self, row, col, res):
        """ Set cell with index (row, col) to be full """
        
        self._cells[row][col] = res

    def is_empty(self, row, col):
        """ Checks whether cell with index (row, col) is empty """
        
        return self._cells[row][col] == EMPTY

    def is_gap(self, row, col):
        """ Checks whetehr cell with indxex (row, col) is a gap """
        
        return self._cells[row][col] == GAP

    def insert_gaps(self, cols):
        """ Inserts a gaps into a column of the template grid """
        
        for col in cols:
            for row in range(self._grid_height):
                if col >= self.get_seq_start(row) and col < self.get_seq_end(row):
                    self._cells[row].insert(col, GAP)
                else:
                    self._cells[row].insert(col, EMPTY)
            
            self._grid_width += 1

    def insert_gaps_row(self, cols, row):
        """ Intert gaps into cols only for certain row"""
        for col in cols:
            if col >= self.get_seq_start(row) and col < self.get_seq_end(row):
                self._cells[row].insert(col, GAP)
            else:
                self._cells[row].insert(col, EMPTY)
            # NOTE: grid_with should not be changed after every row is updated.    
            #self._grid_width += 1

    def clean_trail_empty(self):
        """ Remove all trailing EMPTY and pad grid to same width"""
        # first find out the max length (exluding trailing EMPTY)
        max_width = 0
        for row in range(self._grid_height):
            for i in range(len(self._cells[row])-1, -1, -1):
                if self._cells[row][i] != EMPTY:
                    break
            if i+1 > max_width:
                max_width = i+1
                
        # delete excessive EMPTY        
        for row in range(self._grid_height):
            del self._cells[row][max_width:]

        # then pad all rows to the same length
        [self._cells[row].append( EMPTY * (max_width-len(self._cells[row])) ) \
                                 for row in range(self._grid_height) if len(self._cells[row]) < max_width]
        self._grid_width = max_width
        return
    
    def remove_gaps(self, keep_width=True): # hzhu add keep_width option
        """ Removes all gaps from the grid. """

        for row in range(self.get_grid_height()):
            not_gap = list()
            for col in range(self.get_grid_width()):
                if not self.is_gap(row, col):
                    not_gap.append(col)

            self._cells[row] = [ self._cells[row][col] for col in not_gap ]

            if keep_width:  # hzhu only pad to original width if desired
                for del_pos in range(self._grid_width - len(not_gap)):
                    self._cells[row].append(EMPTY)

        if not keep_width: # hzhu if width is not kept, make sure width is consistent
            self.clean_trail_empty()
                    
        return
                

class QueryGrid(Grid):

    def __init__(self, grid_height, grid_width):
        Grid.__init__(self, grid_height, grid_width)
    
    def get_query_start(self, row):
        """ Returns the query start position """
        return self.get_seq_start(row) + 1

    def get_query_end(self, row):
        """ Returns the query end postion """
        
        return self.get_seq_end(row) - len(self.get_gaps(row))

    def get_col_residue(self, col):
        """ Tries to find a the query residue in a given column. Used by derive_global_seq() to
        identify the global query sequence """
        
        for row in range(self.get_grid_height()):
            if not self.is_empty(row, col):
                return self._cells[row][col]

        return GAP

class TemplateGrid(Grid):

    def __init__(self, grid_height, grid_width):
        Grid.__init__(self, grid_height, grid_width)

        self._start = list()
        self._end = list()
        self._pdb_code = list()
        self._chain = list()
        self._organism = list()
        self._resolution = list()
        
    def display(self):
        """ Return multi-line string represenation for grid """
        
        ans = ''
        for row in range(self._grid_height):
            ans += '>P1;{p}\nstructure:{p}:{s}:{c}:{e}:{c}::{o}:{r}:\n{a}*\n'.format(
                p = self._pdb_code[row],
                s = add_white_space_end(self.get_template_start(row), 4),
                e = add_white_space_end(self.get_template_end(row), 4),
                c = self._chain[row],
                o = self._organism[row],
                r = self._resolution[row], 
                a = ''.join(self._cells[row]).replace(EMPTY, GAP).replace('#', GAP))

        return ans

    def debug(self, row):
        """ Return multi-line string represenation for grid, for debugging purposes """

        ans = '{p}\nInternal: {s}, {e} Query: {qs}, {qe} Gaps ({g1}): {g2}\n{seq}\n'.format(
            p = self._pdb_code[row],
            s = self.get_seq_start(row),
            e = self.get_seq_end(row),
            qs = self.get_template_start(row),
            qe = self.get_template_end(row),
            g1 = len(self.get_gaps(row)),
            g2 = ', '.join([str(gap) for gap in self.get_gaps(row)]),
            seq = ''.join(self._cells[row]))

        return ans 

    def set_metadata(self, row, start, end, pdb_code, chain, organism, resolution):
        """ Used by create_template_grid() to setup metadata of pir template """
        
        self._start.append(start)
        self._end.append(end)
        self._pdb_code.append(pdb_code)
        self._chain.append(chain)
        self._organism.append(organism)
        self._resolution.append(resolution)

    def set_map(self, row, start, end):
        
        self._start[row] = start
        self._end[row] = end

    def get_template_start(self, row):
        """ Returns the template start position """
        
        return self._start[row]

    def get_template_end(self, row):
        """ Return sthe template end position """
        
        return self._end[row]

    def del_row(self, row):
        """ Removes a complete template entry from the grid """
        
        del self._cells[row]
        del self._start[row]
        del self._end[row]
        del self._pdb_code[row]
        del self._chain[row]
        del self._organism[row]
        del self._resolution[row]
        self._grid_height -= 1

# Helper functions

def add_white_space_end(string, length):
    """ Adds whitespaces to a string until it has the wished length"""
    
    edited_string = str(string)
    
    if len(edited_string) >= length:
        return string
    else:
        while len(edited_string) != length:
            edited_string += ' '
    
    return edited_string

def convert_aa_code(three_letter, convert):
    """
    Assumes a string that contains a three letter aminoacid code and
    returns the corresponding one letter code.
    """    
    
    aa_code = {
        'CYS': 'C', 
        'ASP': 'D', 
        'SER': 'S', 
        'GLN': 'Q', 
        'LYS': 'K',
        'ILE': 'I', 
        'PRO': 'P', 
        'THR': 'T', 
        'PHE': 'F', 
        'ASN': 'N', 
        'GLY': 'G', 
        'HIS': 'H', 
        'LEU': 'L', 
        'ARG': 'R', 
        'TRP': 'W', 
        'ALA': 'A', 
        'VAL': 'V', 
        'GLU': 'E', 
        'TYR': 'Y', 
        'MET': 'M',
      } 

    non_canonical = {
        'MSE': 1,
        'HYP': 2,
        'MLY': 3,
        'SEP': 4,
        'TPO': 5,
        'CSO': 6,
        'PTR': 7,
        'KCX': 8,
        'CME': 9,
        'CSD': 10,
        'CAS': 11,
        'MLE': 12,
        'DAL': 13,
        'CGU': 14,
        'DLE': 15,
        'FME': 16,
        'DVA': 17,
        'OCS': 18,
        'DPR': 19,
        'MVA': 20,
        'TYS': 21,
        'M3L': 22,
        'SMC': 23,
        'ALY': 24,
        'CSX': 25,
        'DCY': 26,
        'NLE': 27,
        'DGL': 28,
        'DSN': 29,
        'CSS': 30,
        'DLY': 31,
        'MLZ': 32,
        'DPN': 33,
        'DAR': 34,
        'PHI': 35,
        'IAS': 36,
        'DAS': 37,
        'HIC': 38,
        'MP8': 39,
        'DTH': 40,
        'DIL': 41,
        'MEN': 42,
        'DTY': 43,
        'CXM': 44,
        'DGN': 45,
        'DTR': 46,
        'SAC': 47,
        'DSG': 48,
        'MME': 49,
        'MAA': 50,
        'YOF': 51,
        'FP9': 52,
        'FVA': 53,
        'MLU': 54,
        'OMY': 55,
        'FGA': 56,
        'MEA': 57,
        'CMH': 58,
        'DHI': 59,
        'SEC': 60,
        'OMZ': 61,
        'SCY': 62,
        'MHO': 63,
        'MED': 64,
        'CAF': 65,
        'NIY': 66,
        'OAS': 67,
        'SCH': 68,
        'MK8': 69,
        'SME': 70,
        'LYZ': 71
    }

    if three_letter in aa_code.keys():
        return aa_code[three_letter]
    elif convert and (three_letter in non_canonical.keys()):
        return non_canonical[three_letter]
    else:
        return '-'


def get_query_name(hhr_file):

    with open(hhr_file) as fh:
        for line in fh:
            if line.startswith('Query'):
                # match the PDB Code
                m = re.search('(\d[A-Z0-9]{3})_(\S)', line) 

                if m: 
                    pdb_code = m.group(1)
                    chain = m.group(2)
                else: 
                    pdb_code = 'UKNP'
                    chain = 'A'
                    # raise ValueError('Input HHR-File Does not seem to be a PDB-Structure')

                break

    return pdb_code, chain

def get_cif_files(folder):
    """ Gets all cif files located in folder. """
    
    return glob(os.path.join(folder, '*.cif'))

def open_cif(cif_file):
    """ Assumes a mmCif file and returns a data block used for subsequent procedures """
    # The "usual" procedure to open a mmCIF with pdbX/mmCIF
    with open(cif_file) as cif_fh:
        data = []
        reader = PdbxReader(cif_fh)
        reader.read(data)

    block = data[0]
    
    return block

def get_pdb_entry_id(block):
    """ Extracts the PDB entry information of a cif file and returns it as a string """

    entry = block.getObj('entry')
    entry_id = entry.getValue('id')
    
    return entry_id


def template_id_to_pdb(template_id):
    """
    Extracts PDB ID and chain name from the provided template id
    """
    # match PDBID without chain (8fab, 1a01)
    m = re.match(r'/^(\d[A-Za-z0-9]{3})$', template_id)
    if m:
        return m.group(1).upper(), 'A'
    
    # PDB CODE with chain Identifier
    m = re.match(r'^(\d[A-Za-z0-9]{3})_(\S)$', template_id)
    if m:
        return m.group(1).upper(), m.group(2).upper()

    # Match DALI ID
    m = re.match(r'^(\d[A-Za-z0-9]{3})([A-Za-z0-9]?)_\d+$', template_id)
    if m:
        return m.group(1).upper(), m.group(2).upper()
    
    # No PDB code and chain identified
    return None, None


def create_template_grid(hhr_data):
    """ Creates a template grid """

    total_seq = len(hhr_data)
    templ_max = max( [ hhr.start[0] + len(to_seq(hhr.template_ali)) for hhr in hhr_data ] ) - 1


    template_grid = TemplateGrid(total_seq, templ_max)

    for row, template in enumerate(hhr_data):
        seq_start = template.start[0] - 1
        templatealignment = to_seq(template.template_ali)
        seq_end = seq_start + len(templatealignment)

        # Load Meta Data
        start = template.start[1]
        end = template.end[1]

        # Get pdb_code and chain identifier of template
        pdb_code, chain =  template_id_to_pdb(template.template_id)

        m = re.search("(\d+.\d+)A", template.template_info) # try to extract resolution of the structure
        
        if m: 
            resolution = m.group(1)
        else: 
            resolution = ""

        m = re.search("\{(.*)\}", template.template_info) # try to extract the organism
        if m: 
            organism = m.group(1).replace(":", " ") # make sure that no colons are in the organism
        else: 
            organism = ""

        template_grid.set_metadata(row, start, end, pdb_code, chain, organism, resolution)

        # Write sequence into the grid
        for pos, col in enumerate(range(seq_start, seq_end)):
            template_grid.set_cell(row, col, templatealignment[pos])

    return template_grid


def to_seq(ali):
    if isinstance(ali, list):
        return ''.join(ali)
    else:
        return ali
    

def create_query_grid(hhr_data):
    """ Creates a Query Grid """
    
    total_seq = len(hhr_data)
    query_max = max( [ hhr.start[0] + len(to_seq(hhr.query_ali)) for hhr in hhr_data ] ) - 1

    query_grid = QueryGrid(total_seq, query_max)

    for row, query in enumerate(hhr_data):

        queryalignment = to_seq(query.query_ali)
        query_start = query.start[0] - 1
        query_end = query_start + len(queryalignment)

        for pos, col in enumerate(range(query_start, query_end)):
            if queryalignment[pos] not in ['Z', 'U', 'O', 'J', 'X', 'B']: # CAUTION

                query_grid.set_cell(row, col, queryalignment[pos])

    return query_grid

def create_gapless_grid(grid):
    """ Returns a gapless grid """

    gapless = deepcopy(grid)
    gapless.remove_gaps(keep_width=False)  # hzhu:  shrink grid

    return gapless

def process_query_grid(query_grid, gapless_grid):
    """ Processes a query grid sucht that it contains all gaps
    """
    gaplist = query_grid.get_gap_list()
    off_set = 0
    
    for g in gaplist:
        gapless_grid.insert_gaps([ p + off_set for p in range(g.open_pos, g.open_pos+g.size) ])
        off_set += g.size
        
    return gapless_grid

def derive_global_seq(processed_query_grid, query_name, query_chain):

    global_seq = list()

    for col in range(processed_query_grid.get_grid_width()):
        global_seq.append(processed_query_grid.get_col_residue(col))

    # this is the query entry
    header = '>P1;{q}\nsequence:{q}:1    :{c}:{l}  :{c}::::\n'.format(
        q = query_name, 
        l = len(global_seq),
        c = query_chain)

    return header + ''.join(global_seq) + '*'

def process_template_grid(query_grid, template_grid):
    """ Insertes Gaps into the template grid
        Only add gaps from **other** query_grids into template grid (NOT gapless)
    """
    gaplist = query_grid.get_gap_list()  # use this to keep the offset
    
    for row in range(template_grid.get_grid_height()):
        # do NOT consider gaps in current query row         
        gaplist_row = query_grid.get_gaps_ref_gapless(row)
        gapdict_row = dict(zip([g.open_pos for g in gaplist_row],
                               [g.size     for g in gaplist_row]))
        off_set = 0
        for g in gaplist:
            # if there is a gap with same opening position in the current row,
            # only consider g if it is larger than the on in the current row
            if g.open_pos in gapdict_row:
                if g.size > gapdict_row[g.open_pos]: 
                    template_grid.insert_gaps_row([ p + off_set for p in range(g.open_pos,
                                                                              g.open_pos+g.size-gapdict_row[g.open_pos]) ], row)
            else:
                template_grid.insert_gaps_row([ p + off_set for p in range(g.open_pos, g.open_pos+g.size) ], row)
                 
            off_set += g.size  # even if the gaps are not inserted, the offset should be adjusted

    template_grid.clean_trail_empty()  # clean the redundant trailing EMPTY char
    
    return template_grid

def compare_with_cifs(template_grid, folder, output_path, convert, threshold):
    """
    Compare the PIR Alignment with Atomsection of a mmCIF file. To make the ATOM-Section of 
    a mmCIF file compatible with MODELLER, each residue has in the ATOM-Section has to match
    corresponding positions in the PIR-Alignment
    """

    # glob the mmCif files from given directory and map the PDB identifier to the path
    cif_files = glob.glob(os.path.join(folder, '*.cif'))
    cif_paths = { path.split('/')[-1].split('.')[0].upper() : path for path in cif_files }
    cif_edits = dict()
    

    # create the path where renumbered cifs are saved to
    if not os.path.exists(output_path):
        os.mkdir(output_path)
    
    # if the cif does not contain any residue of the por alignment we delete it
    del_row = list()

    for row in range(template_grid.get_grid_height()):
        # get the pdb code and strand id from the current template
        pdb_code = template_grid._pdb_code[row]
        chain = template_grid._chain[row]  # hhr users pdb chain ID


        # load mmCif file accordingly
        if pdb_code in cif_edits.keys():
            block = cif_edits[pdb_code]
        else:
            try:
                block = open_cif(cif_paths[pdb_code])
            except KeyError:
                del_row.append(row)
                print ('! Did not find the mmCIF file for {pdb}. Removing it from the alignment.'.format(
                    pdb = pdb_code))
                continue

        # Create a mapping of the atom site
        atom_site = block.getObj('atom_site')

        ########################################################################
        ## Get the mapping of the residues in the atom section                ##
        ########################################################################

        cif_seq = dict()
        # For the case that we have to rename a chain
        cif_chains = set([]) 

        # Iterate through the atomsection of the cif file
        for atom_row in range(0, atom_site.getRowCount()):
            
            try:
                if atom_site.getValue('label_comp_id', atom_row) == 'HOH':
                    continue
                cif_chain = atom_site.getValue('label_asym_id', atom_row)
                pdb_chain = atom_site.getValue('auth_asym_id', atom_row)   # use PDB chain ID
            except IndexError:
                pass

            cif_chains.add(cif_chain)

            # We do not care about the residues apart from the chain
            #if cif_chain != chain: # hzhu
            if pdb_chain != chain:  # hhr uses PDB chain, not the cif chain! hzhu
                continue     
            # and update the chain id from pdb_chain to cif_chain
            if atom_site.getValue('group_PDB', atom_row).startswith('ATOM'):  # hzhu in case HETATM ruins ch id
                template_grid._chain[row] = cif_chain
                
            # get the residue and the residue number 
            try:
                res_num = int(atom_site.getValue("label_seq_id", atom_row))
            except ValueError:
                continue

            residue = atom_site.getValue('label_comp_id', atom_row)
            residue = convert_aa_code(residue, convert)         


            if res_num not in cif_seq.keys():
                cif_seq[res_num] = residue
            elif res_num in cif_seq.keys() and cif_seq[res_num] == residue:
                continue
            elif res_num in cif_seq.keys() and cif_seq[res_num] != residue:
                cif_seq[res_num] = '-'
                
                if DEBUG_MODE:
                    print ('! {p} {c}: mmCIF contains a residue position that is assigned {cr} to two residues. Removing it.'.format(
                        p = pdb_code,
                        c = chain,
                        cr = res_num))

        ########################################################################
        ## Rename chain if necessary                                          ##
        ########################################################################

        chain_idx = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

        if len(template_grid._chain[row]) != 1:
            i = 0
            new_chain = 0
            
            while i < len(chain_idx):
                if chain_idx[i] in cif_chains:
                    if DEBUG_MODE:
                        print ('! {p} {c}: Chain identifier {i} is already taken.'.format(
                            p = pdb_code,
                            c = chain,
                            i = chain_idx[i]))
                    i += 1
                else:
                    new_chain = chain_idx[i]
                    break

            if new_chain == 0:
                if DEBUG_MODE:
                    print ('! {p} {c}: Could not use {p}. The chain identifier {c} is not compatible with MODELLER (2 letters) and could not be renanmed.'.format(
                        p = pdb_code, 
                        c = chain))
                
                del_row.append(row)
                continue

            if new_chain != 0:
                print ('Selected new chain name {c}'.format(c = new_chain))

            #TODO

        ########################################################################
        ## Compare cif positions with the atom positions                      ##
        ########################################################################

        del_pos = list()
        mod_pos = dict()
        mapping = dict()

        for pos_cif, pos_tem in zip(range(template_grid.get_template_start(row), 
            template_grid.get_template_end(row) + 1), template_grid.get_seq_indeces(row)):

            res_tem = template_grid.get_cell(row, pos_tem)

            try:
                res_cif = cif_seq[pos_cif]
            except KeyError:
                res_cif = -1
                

            match = True if res_tem == res_cif else False

            if not match:
                if res_cif == 1 and res_tem == 'M':
                    mod_pos[pos_cif] = 1
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'M')

                elif res_cif == 2 and res_tem == 'P':

                    mod_pos[pos_cif] = 2
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'P')

                elif res_cif == 3 and res_tem == 'K':

                    mod_pos[pos_cif] = 3
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'K')

                elif res_cif == 4 and res_tem == 'S':

                    mod_pos[pos_cif] = 4
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'S')

                elif res_cif == 5 and res_tem == 'T':

                    mod_pos[pos_cif] = 5
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'T')

                elif res_cif == 6 and res_tem == 'C':

                    mod_pos[pos_cif] = 6
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'C')

                elif res_cif == 7 and res_tem == 'Y':

                    mod_pos[pos_cif] = 7
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'Y')

                elif res_cif == 8 and res_tem == 'K':

                    mod_pos[pos_cif] = 8
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'K')

                elif res_cif == 9 and res_tem == 'C':

                    mod_pos[pos_cif] = 9
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'C')

                elif res_cif == 10 and res_tem == 'A':

                    mod_pos[pos_cif] = 10
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'A')

                elif res_cif == 11 and res_tem == 'C':

                    mod_pos[pos_cif] = 11
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'C')

                elif res_cif == 12 and res_tem == 'L':

                    mod_pos[pos_cif] = 12
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'L')

                elif res_cif == 13 and res_tem == 'A':

                    mod_pos[pos_cif] = 13
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'A')

                elif res_cif == 14 and res_tem == 'E':

                    mod_pos[pos_cif] = 14
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'E')

                elif res_cif == 15 and res_tem == 'L':

                    mod_pos[pos_cif] = 15
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'L')

                elif res_cif == 16 and res_tem == 'M':

                    mod_pos[pos_cif] = 16
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'M')

                elif res_cif == 17 and res_tem == 'V':

                    mod_pos[pos_cif] = 17
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'V')

                elif res_cif == 18 and res_tem == 'C':

                    mod_pos[pos_cif] = 18
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'C')

                elif res_cif == 19 and res_tem == 'P':

                    mod_pos[pos_cif] = 19
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'P')

                elif res_cif == 20 and res_tem == 'V':

                    mod_pos[pos_cif] = 20
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'V')

                elif res_cif == 21 and res_tem == 'Y':

                    mod_pos[pos_cif] = 21
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'Y')

                elif res_cif == 22 and res_tem == 'K':

                    mod_pos[pos_cif] = 22
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'K')

                elif res_cif == 23 and res_tem == 'C':

                    mod_pos[pos_cif] = 23
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'C')

                elif res_cif == 24 and res_tem == 'K':

                    mod_pos[pos_cif] = 24
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'K')

                elif res_cif == 25 and res_tem == 'C':

                    mod_pos[pos_cif] = 25
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'C')

                elif res_cif == 26 and res_tem == 'C':

                    mod_pos[pos_cif] = 26
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'C')

                elif res_cif == 27 and res_tem == 'L':

                    mod_pos[pos_cif] = 27
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'L')

                elif res_cif == 28 and res_tem == 'E':

                    mod_pos[pos_cif] = 28
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'E')

                elif res_cif == 29 and res_tem == 'S':

                    mod_pos[pos_cif] = 29
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'S')

                elif res_cif == 30 and res_tem == 'C':

                    mod_pos[pos_cif] = 30
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'C')

                elif res_cif == 31 and res_tem == 'K':

                    mod_pos[pos_cif] = 31
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'K')

                elif res_cif == 32 and res_tem == 'K':

                    mod_pos[pos_cif] = 32
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'K')

                elif res_cif == 33 and res_tem == 'F':

                    mod_pos[pos_cif] = 33
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'F')

                elif res_cif == 34 and res_tem == 'R':

                    mod_pos[pos_cif] = 34
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'R')

                elif res_cif == 35 and res_tem == 'F':

                    mod_pos[pos_cif] = 35
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'F')

                elif res_cif == 36 and res_tem == 'D':

                    mod_pos[pos_cif] = 36
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'D')

                elif res_cif == 37 and res_tem == 'D':

                    mod_pos[pos_cif] = 37
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'D')

                elif res_cif == 38 and res_tem == 'H':

                    mod_pos[pos_cif] = 38
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'H')

                elif res_cif == 39 and res_tem == 'P':

                    mod_pos[pos_cif] = 39
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'P')

                elif res_cif == 40 and res_tem == 'T':

                    mod_pos[pos_cif] = 40
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'T')

                elif res_cif == 41 and res_tem == 'I':

                    mod_pos[pos_cif] = 41
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'I')

                elif res_cif == 42 and res_tem == 'N':

                    mod_pos[pos_cif] = 42
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'N')

                elif res_cif == 43 and res_tem == 'Y':

                    mod_pos[pos_cif] = 43
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'Y')

                elif res_cif == 44 and res_tem == 'M':

                    mod_pos[pos_cif] = 44
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'M')

                elif res_cif == 45 and res_tem == 'G':

                    mod_pos[pos_cif] = 45
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'G')

                elif res_cif == 46 and res_tem == 'W':

                    mod_pos[pos_cif] = 46
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'W')

                elif res_cif == 47 and res_tem == 'S':

                    mod_pos[pos_cif] = 47
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'S')

                elif res_cif == 48 and res_tem == 'N':

                    mod_pos[pos_cif] = 48
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'N')

                elif res_cif == 49 and res_tem == 'M':

                    mod_pos[pos_cif] = 49
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'M')

                elif res_cif == 50 and res_tem == 'A':

                    mod_pos[pos_cif] = 50
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'A')

                elif res_cif == 51 and res_tem == 'Y':

                    mod_pos[pos_cif] = 51
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'Y')

                elif res_cif == 52 and res_tem == 'P':

                    mod_pos[pos_cif] = 52
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'P')

                elif res_cif == 53 and res_tem == 'V':

                    mod_pos[pos_cif] = 53
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'V')

                elif res_cif == 54 and res_tem == 'L':

                    mod_pos[pos_cif] = 54
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'L')

                elif res_cif == 55 and res_tem == 'Y':

                    mod_pos[pos_cif] = 55
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'Y')

                elif res_cif == 56 and res_tem == 'E':

                    mod_pos[pos_cif] = 56
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'E')
                
                elif res_cif == 57 and res_tem == 'F':

                    mod_pos[pos_cif] = 57
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'F')

                elif res_cif == 58 and res_tem == 'C':

                    mod_pos[pos_cif] = 58
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'C')

                elif res_cif == 59 and res_tem == 'H':

                    mod_pos[pos_cif] = 59
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'H')

                elif res_cif == 60 and res_tem == 'C':

                    mod_pos[pos_cif] = 60
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'C')

                elif res_cif == 61 and res_tem == 'Y':

                    mod_pos[pos_cif] = 61
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'Y')

                elif res_cif == 62 and res_tem == 'C':

                    mod_pos[pos_cif] = 62
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'C')

                elif res_cif == 63 and res_tem == 'M':

                    mod_pos[pos_cif] = 63
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'M')

                elif res_cif == 64 and res_tem == 'M':

                    mod_pos[pos_cif] = 64
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'M')

                elif res_cif == 65 and res_tem == 'C':

                    mod_pos[pos_cif] = 65
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'C')

                elif res_cif == 66 and res_tem == 'Y':

                    mod_pos[pos_cif] = 66
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'Y')

                elif res_cif == 67 and res_tem == 'S':

                    mod_pos[pos_cif] = 67
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'S')

                elif res_cif == 68 and res_tem == 'C':

                    mod_pos[pos_cif] = 68
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'C')

                elif res_cif == 69 and res_tem == 'L':

                    mod_pos[pos_cif] = 69
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'L')

                elif res_cif == 70 and res_tem == 'M':

                    mod_pos[pos_cif] = 70
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'M')

                elif res_cif == 71 and res_tem == 'K':

                    mod_pos[pos_cif] = 71
                    mapping[(pos_tem, res_tem)] = (pos_cif, 'K')

                else:
                    # insert a gap
                    template_grid.set_empty(row, pos_tem)
                    mapping[(pos_tem, res_tem)] = (pos_cif, res_cif)
                    
                    if DEBUG_MODE:
                        print ('! {p} {c}: template pos {pt} ({rt}) does not match cif pos {pc} ({rc}). Replacing with gap.'.format( 
                            p = pdb_code, 
                            c = chain,
                            pt = pos_tem, 
                            rt = res_tem, 
                            pc = pos_cif, 
                            rc = res_cif if res_cif != -1 else 'DNE'))
                    
                    if res_cif != -1:
                        del_pos.append(pos_cif)
            else:
                mapping[(pos_tem, res_tem)] = (pos_cif, res_cif)


        # adjust template start and end positions
        correct_mapping = { key:value for key, value in mapping.items() if key[1] == value[1] }

        try:
            tstart = correct_mapping[sorted(correct_mapping.keys())[0]][0]
            tend = correct_mapping[sorted(correct_mapping.keys())[-1]][0]
            template_grid.set_map(row, tstart, tend)
        except IndexError:
            # This exception handles cases in which all residues were deleted
            if DEBUG_MODE:
                print ('! {p} {c}: Removing {p} from alignment. No residues matched the alignment sequence.'.format( 
                    p = pdb_code, 
                    c = chain))

            del_row.append(row)
            continue

        ########################################################################
        ## Delete rows from the PIR Alignment if the residue ratio is to low  ##
        ########################################################################

        if threshold > 0:
            
            gaps = 0
            res = 0

            for col in range(template_grid.get_grid_width()):
                if template_grid.is_empty(row, col):
                    template_grid.set_gap(row, col)

                if template_grid.is_gap(row, col):
                    gaps += 1
                else:
                    res += 1

            ratio = res/float(gaps + res)
        
            if ratio > threshold:
                print ('! Template {p} successfully passed residue ratio ({r:.2f} / {t}).'.format( 
                    p = pdb_code,
                    r = ratio,
                    t = threshold ))
            else:
                print ('! Template {p} did not passed residue ratio ({r:.2f} / {t}). Removing it from pir Alignment.'.format( 
                    p = pdb_code,
                    r = ratio,
                    t = threshold ))            
            
                if row not in del_row:
                    del_row.append(row)
                    continue

        ########################################################################
        ## Edit cif files                                                     ##
        ########################################################################

        rem_row = list() # verbosity: saves information about removed residues
        mod_row = list() # verbosity: saves information about modified residues
        cha_row = list() # verbosity: saves any other changes 

        for atom_row in reversed(range(0, atom_site.getRowCount())):
            
            try:
                cif_chain = atom_site.getValue('label_asym_id', atom_row)
            except IndexError:
                pass

            # We do not care about the residues apart from the chain
            if cif_chain != chain:
                continue    

            # get the residue number 
            try:
                res_num = int(atom_site.getValue("label_seq_id", atom_row))
            except ValueError:
                continue

            # pdb_PDB_model_num has to be set to 1
            try:
                model_num = int(atom_site.getValue('pdbx_PDB_model_num', atom_row))
            except IndexError:
                model_num = 1 # if we cannot extract, assume that it is alright

            try:
                ins_code = atom_site.getValue('pdbx_PDB_ins_code', atom_row)
            except IndexError:
                ins_code = '?' # assume it has no insertion code

            group_PDB = atom_site.getValue('group_PDB', atom_row)
            residue = atom_site.getValue('label_comp_id', atom_row)
            residue = convert_aa_code(residue, convert)         

            # MODELLER accepts only structures if pdbx_PDB_model_num is set to 1   
            if model_num != 1:

                if (res_num, residue, 'model_num') not in cha_row:
                    cha_row.append((res_num, residue, 'model_num'))             
                
                atom_site.setValue(1, "pdbx_PDB_model_num", atom_row)

            if ins_code != '?':
                
                if (res_num, residue, 'ins_code') not in cha_row:
                    cha_row.append((res_num, residue, 'ins_code'))
                
                atom_site.setValue('?', "pdbx_PDB_ins_code", atom_row)

            if group_PDB != 'ATOM':

                if (res_num, residue, 'group_PDB') not in cha_row:
                    cha_row.append((res_num, residue, 'group_PDB'))

                atom_site.setValue('ATOM', 'group_PDB', atom_row)               

            ########################################################################
            ## Delete residues                                                    ##
            ########################################################################

            if res_num in del_pos:
                if (res_num, residue) not in rem_row:
                    rem_row.append((res_num, residue))
                
                atom_site.removeRow(atom_row)

            ########################################################################
            ## Modify residues                                                    ##
            ########################################################################

            if res_num in mod_pos.keys():
                
                # Get the data
                type_symbol = atom_site.getValue('type_symbol', atom_row)
                label_atom_id = atom_site.getValue('label_atom_id', atom_row)
                auth_atom_id = atom_site.getValue('auth_atom_id', atom_row)

                if mod_pos[res_num] == 1: # try to convert MSE to M 
                                        
                    atom_site.setValue('MET', 'label_comp_id', atom_row)
                    
                    try:
                        atom_site.setValue('MET', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if type_symbol == 'SE':
                        atom_site.setValue('S', 'type_symbol', atom_row)
                    if label_atom_id == 'SE':
                        atom_site.setValue('S', 'label_atom_id', atom_row)
                    if auth_atom_id == 'SE':
                        atom_site.setValue('S', 'auth_atom_id', atom_row)

                    if (res_num, residue, 'MSE -> MET') not in mod_row:
                        mod_row.append((res_num, residue, 'MSE -> MET'))

                elif mod_pos[res_num] == 2: # try to convert HYP to PRO
                    # apparently it is enough to rename the label_comp_id to PRO to get 
                    # MODELLER working with Hydroxyprolines (HYP)

                    atom_site.setValue('PRO', 'label_comp_id', atom_row)
                    
                    try:
                        atom_site.setValue('PRO', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'HYP -> PRO') not in mod_row:
                        mod_row.append((res_num, residue, 'HYP -> PRO'))

                elif mod_pos[res_num] == 3: # try to convert MLY to LYS

                    atom_site.setValue('LYS', 'label_comp_id', atom_row)
                    
                    try:
                        atom_site.setValue('LYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'MLY -> LYS') not in mod_row:
                        mod_row.append((res_num, residue, 'MLY -> LYS'))

                elif mod_pos[res_num] == 4: # converts Phosphoserine to Serine

                    atom_site.setValue('SER', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('SER', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'SEP -> SER') not in mod_row:
                        mod_row.append((res_num, residue, 'SEP -> SER'))

                elif mod_pos[res_num] == 5: # converts Phosphothreonine to Threonine

                    atom_site.setValue('THR', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('THR', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'TPO -> THR') not in mod_row:
                        mod_row.append((res_num, residue, 'TPO -> THR'))

                elif mod_pos[res_num] == 6: # converts S-HYDROXYCYSTEINE to Cysteine

                    atom_site.setValue('CYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('CYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'CSO -> CYS') not in mod_row:
                        mod_row.append((res_num, residue, 'CSO -> CYS'))

                elif mod_pos[res_num] == 7: # converts O-PHOSPHOTYROSINE to Tyrosine

                    atom_site.setValue('TYR', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('TYR', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'PTR -> TYR') not in mod_row:
                        mod_row.append((res_num, residue, 'PTR -> TYR'))

                elif mod_pos[res_num] == 8: # converts LYSINE NZ-CARBOXYLIC ACID to Lysine

                    atom_site.setValue('LYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('LYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'KCX -> LYS') not in mod_row:
                        mod_row.append((res_num, residue, 'KCX -> LYS'))

                elif mod_pos[res_num] == 9: # converts S,S-(2-HYDROXYETHYL)THIOCYSTEINE to Cysteine

                    atom_site.setValue('CYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('CYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'CME -> CYS') not in mod_row:
                        mod_row.append((res_num, residue, 'CME -> CYS'))

                elif mod_pos[res_num] == 10: # converts 3-SULFINOALANINE to Alanine

                    atom_site.setValue('ALA', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('ALA', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'CSD -> ALA') not in mod_row:
                        mod_row.append((res_num, residue, 'CSD -> ALA'))

                elif mod_pos[res_num] == 11: # converts S-(DIMETHYLARSENIC)CYSTEINE to Cysteine

                    atom_site.setValue('CYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('CYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'CAS -> CYS') not in mod_row:
                        mod_row.append((res_num, residue, 'CAS -> CYS'))

                elif mod_pos[res_num] == 12: # converts N-METHYLLEUCINE (MLE) to Leucine

                    atom_site.setValue('LEU', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('LEU', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'MLE -> LEU') not in mod_row:
                        mod_row.append((res_num, residue, 'MLE -> LEU'))

                elif mod_pos[res_num] == 13: # converts D-ALANINE (DAL) to ALA

                    atom_site.setValue('ALA', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('ALA', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DAL -> ALA') not in mod_row:
                        mod_row.append((res_num, residue, 'DAL -> ALA'))

                elif mod_pos[res_num] == 14: # converts GAMMA-CARBOXY-GLUTAMIC ACID (CGU) to GLU

                    atom_site.setValue('GLU', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('GLU', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'CGU -> GLU') not in mod_row:
                        mod_row.append((res_num, residue, 'CGU -> GLU'))

                elif mod_pos[res_num] == 15: # converts D-LEUCINE (DLE) to LEU

                    atom_site.setValue('LEU', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('LEU', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DLE -> LEU') not in mod_row:
                        mod_row.append((res_num, residue, 'DLE -> LEU'))

                elif mod_pos[res_num] == 16: # converts N-FORMYLMETHIONINE (FME) to MET

                    atom_site.setValue('MET', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('MET', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'FME -> MET') not in mod_row:
                        mod_row.append((res_num, residue, 'FME -> MET'))

                elif mod_pos[res_num] == 17: # converts D-VAL (DVA) to VAL

                    atom_site.setValue('VAL', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('VAL', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DVA -> VAL') not in mod_row:
                        mod_row.append((res_num, residue, 'DVA -> VAL'))

                elif mod_pos[res_num] == 18: # converts CYSTEINESULFONIC ACID (OCS) to CYS

                    atom_site.setValue('CYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('CYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'OCS -> CYS') not in mod_row:
                        mod_row.append((res_num, residue, 'OCS -> CYS'))

                elif mod_pos[res_num] == 19: # converts D-PROLINE (DPR) to PRO

                    atom_site.setValue('PRO', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('PRO', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DPR -> PRO') not in mod_row:
                        mod_row.append((res_num, residue, 'DPR -> PRO'))

                elif mod_pos[res_num] == 20: # converts N-METHYLVALINE (MVA) to VAL

                    atom_site.setValue('VAL', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('VAL', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'MVA -> VAL') not in mod_row:
                        mod_row.append((res_num, residue, 'MVA -> VAL'))

                elif mod_pos[res_num] == 21: # converts O-SULFO-L-TYROSINE (TYS) to VAL

                    atom_site.setValue('TYR', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('TYR', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'TYS -> TYR') not in mod_row:
                        mod_row.append((res_num, residue, 'TYS -> TYR'))

                elif mod_pos[res_num] == 22: # converts N-TRIMETHYLLYSINE (M3L) to LYS

                    atom_site.setValue('LYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('LYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'M3L -> LYS') not in mod_row:
                        mod_row.append((res_num, residue, 'M3L -> LYS'))

                elif mod_pos[res_num] == 23: # converts S-METHYLCYSTEINE (SMC) to CYS

                    atom_site.setValue('CYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('CYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'SMC -> CYS') not in mod_row:
                        mod_row.append((res_num, residue, 'SMC -> CYS'))

                elif mod_pos[res_num] == 24: # converts N(6)-ACETYLLYSINE (ALY) to LYS

                    atom_site.setValue('LYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('LYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'ALY -> LYS') not in mod_row:
                        mod_row.append((res_num, residue, 'ALY -> LYS'))

                elif mod_pos[res_num] == 25: # converts S-OXY CYSTEINE (CSX) to CYS

                    atom_site.setValue('CYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('CYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'CSX -> CYS') not in mod_row:
                        mod_row.append((res_num, residue, 'CSX -> CYS'))

                elif mod_pos[res_num] == 26: # converts D-CYSTEINE (DCY) to CYS

                    atom_site.setValue('CYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('CYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DCY -> CYS') not in mod_row:
                        mod_row.append((res_num, residue, 'DCY -> CYS'))

                elif mod_pos[res_num] == 27: # converts NORLEUCINE (NLE) to LEU

                    atom_site.setValue('LEU', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('LEU', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'NLE -> LEU') not in mod_row:
                        mod_row.append((res_num, residue, 'NLE -> LEU'))

                elif mod_pos[res_num] == 28: # converts D-GLUTAMIC ACID (DGL) to GLU

                    atom_site.setValue('GLU', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('GLU', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DGL -> GLU') not in mod_row:
                        mod_row.append((res_num, residue, 'DGL -> GLU'))

                elif mod_pos[res_num] == 29: # converts D-SERINE (DSN) to SER

                    atom_site.setValue('SER', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('SER', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DSN -> SER') not in mod_row:
                        mod_row.append((res_num, residue, 'DSN -> SER'))

                elif mod_pos[res_num] == 30: # converts S-MERCAPTOCYSTEINE (CSS) to CYS

                    atom_site.setValue('CYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('CYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'CSS -> CYS') not in mod_row:
                        mod_row.append((res_num, residue, 'CSS -> CYS'))

                elif mod_pos[res_num] == 31: # converts D-LYSINE (DLY) to LYS

                    atom_site.setValue('LYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('LYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DLY -> LYS') not in mod_row:
                        mod_row.append((res_num, residue, 'DLY -> LYS'))

                elif mod_pos[res_num] == 32: # converts N-METHYL-LYSINE (MLZ) to LYS

                    atom_site.setValue('LYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('LYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'MLZ -> LYS') not in mod_row:
                        mod_row.append((res_num, residue, 'MLZ -> LYS'))

                elif mod_pos[res_num] == 33: # converts D-PHENYLALANINE (DPN) to PHE

                    atom_site.setValue('PHE', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('PHE', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DPN -> PHE') not in mod_row:
                        mod_row.append((res_num, residue, 'DPN -> PHE'))

                elif mod_pos[res_num] == 34: # converts D-ARGININE (DAR) to ARG

                    atom_site.setValue('ARG', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('ARG', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DAR -> ARG') not in mod_row:
                        mod_row.append((res_num, residue, 'DAR -> ARG'))

                elif mod_pos[res_num] == 35: # converts IODO-PHENYLALANINE (PHI) to PHE

                    atom_site.setValue('PHE', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('PHE', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'PHI -> PHE') not in mod_row:
                        mod_row.append((res_num, residue, 'PHI -> PHE'))

                elif mod_pos[res_num] == 36: # converts BETA-L-ASPARTIC ACID (IAS) to ASP

                    atom_site.setValue('ASP', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('ASP', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'IAS -> ASP') not in mod_row:
                        mod_row.append((res_num, residue, 'IAS -> ASP'))

                elif mod_pos[res_num] == 37: # converts D-ASPARTIC ACID (DAS) to ASP

                    atom_site.setValue('ASP', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('ASP', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DAS -> ASP') not in mod_row:
                        mod_row.append((res_num, residue, 'DAS -> ASP'))

                elif mod_pos[res_num] == 38: # converts 4-METHYL-HISTIDINE (HIC) to HIS

                    atom_site.setValue('HIS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('HIS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'HIC -> HIS') not in mod_row:
                        mod_row.append((res_num, residue, 'HIC -> HIS'))

                elif mod_pos[res_num] == 39: # converts (4R)-4-methyl-L-proline (MP8) to PRO

                    atom_site.setValue('PRO', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('PRO', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'MP8 -> PRO') not in mod_row:
                        mod_row.append((res_num, residue, 'MP8 -> PRO'))

                elif mod_pos[res_num] == 40: # converts D-THREONINE (DTH) to THR

                    atom_site.setValue('THR', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('THR', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DTH -> THR') not in mod_row:
                        mod_row.append((res_num, residue, 'DTH -> THR'))

                elif mod_pos[res_num] == 41: # converts D-ISOLEUCINE (DIL) to ILE

                    atom_site.setValue('ILE', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('ILE', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DIL -> ILE') not in mod_row:
                        mod_row.append((res_num, residue, 'DIL -> ILE'))

                elif mod_pos[res_num] == 42: # converts N-METHYL ASPARAGINE (MEN) to ASN

                    atom_site.setValue('ASN', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('ASN', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'MEN -> ASN') not in mod_row:
                        mod_row.append((res_num, residue, 'MEN -> ASN'))

                elif mod_pos[res_num] == 43: # converts D-TYROSINE (DTY) to TYR

                    atom_site.setValue('TYR', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('TYR', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DTY -> TYR') not in mod_row:
                        mod_row.append((res_num, residue, 'DTY -> TYR'))

                elif mod_pos[res_num] == 44: # converts N-CARBOXYMETHIONINE (CXM) to MET

                    atom_site.setValue('MET', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('MET', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'CXM -> MET') not in mod_row:
                        mod_row.append((res_num, residue, 'CXM -> MET'))

                elif mod_pos[res_num] == 45: # converts D-GLUTAMINE (DGN) to MET

                    atom_site.setValue('GLN', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('GLN', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DGN -> GLN') not in mod_row:
                        mod_row.append((res_num, residue, 'DGN -> GLN'))

                elif mod_pos[res_num] == 46: # converts D-TRYPTOPHAN (DTR) to TRP

                    atom_site.setValue('TRP', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('TRP', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DTR -> TRP') not in mod_row:
                        mod_row.append((res_num, residue, 'DTR -> TRP'))

                elif mod_pos[res_num] == 47: # converts N-ACETYL-SERINE (SAC) to SER

                    atom_site.setValue('SER', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('SER', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'SAC -> SER') not in mod_row:
                        mod_row.append((res_num, residue, 'SAC -> SER'))

                elif mod_pos[res_num] == 48: # converts D-ASPARAGINE (DSG) to ASN

                    atom_site.setValue('ASN', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('ASN', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DSG -> ASN') not in mod_row:
                        mod_row.append((res_num, residue, 'DSG -> ASN'))

                elif mod_pos[res_num] == 49: # converts N-METHYL METHIONINE (MME) to MET

                    atom_site.setValue('MET', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('MET', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'MME -> MET') not in mod_row:
                        mod_row.append((res_num, residue, 'MME -> MET'))

                elif mod_pos[res_num] == 50: # converts N-methyl-L-alanine (MAA) to ALA

                    atom_site.setValue('ALA', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('ALA', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'MAA -> ALA') not in mod_row:
                        mod_row.append((res_num, residue, 'MAA -> ALA'))

                elif mod_pos[res_num] == 51: # converts 3-FLUOROTYROSINE (YOF) to TYR

                    atom_site.setValue('TYR', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('TYR', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'YOF -> TYR') not in mod_row:
                        mod_row.append((res_num, residue, 'YOF -> TYR'))

                elif mod_pos[res_num] == 52: # converts (4R)-4-fluoro-L-proline (FP9) to PRO

                    atom_site.setValue('PRO', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('PRO', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'FP9 -> PRO') not in mod_row:
                        mod_row.append((res_num, residue, 'FP9 -> PRO'))

                elif mod_pos[res_num] == 53: # converts N-formyl-L-valine (FVA) to VAL

                    atom_site.setValue('VAL', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('VAL', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'FVA -> VAL') not in mod_row:
                        mod_row.append((res_num, residue, 'FVA -> VAL'))

                elif mod_pos[res_num] == 54: # converts N-methyl-D-leucine (MLU) to LEU

                    atom_site.setValue('LEU', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('LEU', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'MLU -> LEU') not in mod_row:
                        mod_row.append((res_num, residue, 'MLU -> LEU'))

                elif mod_pos[res_num] == 55: # converts (betaR)-3-chloro-beta-hydroxy-L-tyrosine (OMY) to TYR

                    atom_site.setValue('TYR', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('TYR', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'OMY -> TYR') not in mod_row:
                        mod_row.append((res_num, residue, 'OMY -> TYR'))

                elif mod_pos[res_num] == 56: # converts GAMMA-D-GLUTAMIC ACID (FGA) to GLU

                    atom_site.setValue('GLU', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('GLU', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'FGA -> GLU') not in mod_row:
                        mod_row.append((res_num, residue, 'FGA -> GLU'))

                elif mod_pos[res_num] == 57: # converts N-METHYLPHENYLALANINE (MEA) to PHE

                    atom_site.setValue('PHE', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('PHE', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'MEA -> PHE') not in mod_row:
                        mod_row.append((res_num, residue, 'MEA -> PHE'))

                elif mod_pos[res_num] == 58: # converts S-(METHYLMERCURY)-L-CYSTEINE (CMH) to CYS

                    atom_site.setValue('CYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('CYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'CMH -> CYS') not in mod_row:
                        mod_row.append((res_num, residue, 'CMH -> CYS'))

                elif mod_pos[res_num] == 59: # converts D-HISTIDINE (DHI) to HIS

                    atom_site.setValue('HIS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('HIS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'DHI -> HIS') not in mod_row:
                        mod_row.append((res_num, residue, 'DHI -> HIS'))

                elif mod_pos[res_num] == 60: # converts SELENOCYSTEINE (SEC) to CYS

                    atom_site.setValue('CYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('CYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'SEC -> CYS') not in mod_row:
                        mod_row.append((res_num, residue, 'SEC -> CYS'))

                elif mod_pos[res_num] == 61: # converts (betaR)-3-CHLORO-BETA-HYDROXY-D-TYROSINE (OMZ) to TYR

                    atom_site.setValue('TYR', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('TYR', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'OMZ -> TYR') not in mod_row:
                        mod_row.append((res_num, residue, 'OMZ -> TYR'))

                elif mod_pos[res_num] == 62: # converts S-ACETYL-CYSTEINE (SCY) to CYS

                    atom_site.setValue('CYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('CYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'SCY -> CYS') not in mod_row:
                        mod_row.append((res_num, residue, 'SCY -> CYS'))

                elif mod_pos[res_num] == 63: # converts S-OXYMETHIONINE (MHO) to MET

                    atom_site.setValue('MET', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('MET', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'MHO -> MET') not in mod_row:
                        mod_row.append((res_num, residue, 'MHO -> MET'))

                elif mod_pos[res_num] == 64: # converts D-METHIONINE (MED) to MET

                    atom_site.setValue('MET', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('MET', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'MED -> MET') not in mod_row:
                        mod_row.append((res_num, residue, 'MED -> MET'))

                elif mod_pos[res_num] == 65: # converts S-DIMETHYLARSINOYL-CYSTEINE (CAF) to CYS

                    atom_site.setValue('CYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('CYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'CAF -> CYS') not in mod_row:
                        mod_row.append((res_num, residue, 'CAF -> CYS'))

                elif mod_pos[res_num] == 66: # converts META-NITRO-TYROSINE (NIY) to TYR

                    atom_site.setValue('TYR', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('TYR', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'NIY -> TYR') not in mod_row:
                        mod_row.append((res_num, residue, 'NIY -> TYR'))

                elif mod_pos[res_num] == 67: # converts O-ACETYLSERINE (OAS) to SER

                    atom_site.setValue('SER', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('SER', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'OAS -> SER') not in mod_row:
                        mod_row.append((res_num, residue, 'OAS -> SER'))

                elif mod_pos[res_num] == 68: # converts S-METHYL-THIO-CYSTEINE (SCH) to CYS

                    atom_site.setValue('CYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('CYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'SCH -> CYS') not in mod_row:
                        mod_row.append((res_num, residue, 'SCH -> CYS'))

                elif mod_pos[res_num] == 69: # converts 2-methyl-L-norleucine (MK8) to LEU

                    atom_site.setValue('LEU', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('LEU', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'MK8 -> LEU') not in mod_row:
                        mod_row.append((res_num, residue, 'MK8 -> LEU'))

                elif mod_pos[res_num] == 70: # converts METHIONINE SULFOXIDE (SME) to MET

                    atom_site.setValue('MET', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('MET', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'SME -> MET') not in mod_row:
                        mod_row.append((res_num, residue, 'SME -> MET'))

                elif mod_pos[res_num] == 71: # converts 5-HYDROXYLYSINE (LYZ) to LYS

                    atom_site.setValue('LYS', 'label_comp_id', atom_row)

                    try:
                        atom_site.setValue('LYS', 'auth_comp_id', atom_row)
                    except IndexError:
                        pass

                    if (res_num, residue, 'LYZ -> LYS') not in mod_row:
                        mod_row.append((res_num, residue, 'LYZ -> LYS'))

        ########################################################################
        ## Notify user about modification made to cif data                    ##
        ########################################################################

        if DEBUG_MODE:
            mod_model_num = len([ msg for msg in cha_row if msg[2] == 'model_num' ])
            mod_ins_code = len([ msg for msg in cha_row if msg[2] == 'ins_code' ])
            mod_group_PDB = len([ msg for msg in cha_row if msg[2] == 'group_PDB' ])
            
            if mod_model_num != 0:
                print ('! {p} {c}: modified atom_site.pdbx_PDB_model_num for {cr} residues to 1.'.format(
                        p = pdb_code,
                        c = chain,
                        cr = mod_model_num))

            if mod_ins_code != 0:
                print ('! {p} {c}: modified atom_site.pdbx_PDB_ins_code for {cr} residues to "?".'.format(
                        p = pdb_code,
                        c = chain,
                        cr = mod_ins_code))

            if mod_group_PDB != 0:
                print ('! {p} {c}: modified atom_site.group_PDB for {cr} residues to "ATOM".'.format(
                        p = pdb_code,
                        c = chain,
                        cr = mod_group_PDB))

            for residue in reversed(mod_row):
                print ('! {p} {c}: modified cif pos {cr} ({nr}).'.format(
                    p = pdb_code, 
                    c = chain,
                    cr = residue[0],
                    ca = residue[1],
                    nr = residue[2]))


            for residue in reversed(rem_row):
                print ('! {p} {c}: removed cif pos {cr} ({ca})'.format(
                    p = pdb_code, 
                    c = chain,
                    cr = residue[0],
                    ca = residue[1]))

        cif_edits[pdb_code] = block

    # write modified pir to disk
    for pdb_code in cif_edits:
        out = open(os.path.join(output_path, pdb_code + '.cif'), 'w')
        writer = PdbxWriter(out)
        writer.writeContainer(cif_edits[pdb_code])

    # Delete missing entries from the last template sequence to the first
    for row in reversed(del_row):
        template_grid.del_row(row)

    return template_grid

def remove_self_alignment(template_grid, query_name):
    """ Removes a self alignment from the final pir alignment to prevent clashes with MODELLER """
    
    to_delete = list()
    
    for row in range(template_grid.get_grid_height()):
        if template_grid._pdb_code[row] == query_name:
            to_delete.append(row)

    for row in reversed(to_delete):
        template_grid.del_row(row)

    return True

def write_to_file(line_list, fname):
    """ Writes the final pir file """
    
    with open(fname, 'w+') as fout:
        for line in line_list:
            fout.write(line + "\n")

def arg():
    import argparse
    description = """Creates a MODELLER alignment (*.pir) from a HHSearch results file (*.hhr)."""
    epilog= '2016 Harald Voehringer.'
    # Initiate a ArgumentParser Class
    parser = argparse.ArgumentParser(description = description, epilog = epilog)
    
    # Call add_options to the parser
    parser.add_argument('input', help = 'results file from HHsearch with hit list and alignment', metavar = 'FILE')    
    parser.add_argument('cifs', help = 'path to the folder containing cif files', metavar = 'DIR')
    parser.add_argument('pir', help = 'output file (PIR-formatted multiple alignment)', metavar = 'FILE')
    parser.add_argument('output', help = 'path to the folder where modified cif files should be written to', metavar = 'DIR')

    parser.add_argument('-v', '--verbose', action = 'store_true', help = 'verbose mode')
    parser.add_argument('-m', nargs = '+', help = 'pick hits with specified indices (e.g. -m 2 5)', metavar = 'INT')
    parser.add_argument('-e', type = float, help = 'maximum E-Value threshold (e.g. -e 0.001)', metavar = 'FLOAT')
    parser.add_argument('-r', type = float, help = 'residue ratio (filter alignments that have contribute at least residues according to the specified ratio).',
        default = 0, metavar = 'FLOAT')

    parser.add_argument('-c', help = 'convert non-canonical residues (default = True)', action = 'store_true', default = True)
    

    return parser


def main():
    import sys
    parser = arg()
    args = parser.parse_args(sys.argv[1:])
    
    global DEBUG_MODE
    
    if args.verbose:
        DEBUG_MODE = True

    query_name, query_chain = get_query_name(args.input)

    data = read_result(args.input)
    selected_templates = list()

    if args.m and not args.e:
        selection = map(lambda x: int(x), args.m)
        print ('Selected templates {st}.'.format(st = ', '.join(args.m)))
        
        for i in selection:
            tmp_info = str(data[i - 1].template_info.split('>')[1])
            print ('{i}: {t}'.format(
                i = i,
                t = tmp_info[0:80]))

            selected_templates.append(data[i - 1])

    elif args.e and not args.m:
        print ('Selected templates satisfying E-val <= {e}'.format(e = args.e))
        
        e_values = { float(j.evalue):i for i, j in enumerate(data) }
        selection = sorted([ val for key, val in e_values.items() if key <= args.e ])

        for i in selection:
            tmp_info = str(data[i - 1].template_info.split('>')[1])
            print ('{i}: {t}'.format(
                i = i + 1,
                t = tmp_info[0:80]))

            selected_templates.append(data[i - 1])

    elif args.m and args.e:
        print ('! Please do not use option -m and -e at the same time ! Exiting.')
        sys.exit()
    else:
        selected_templates = data
        
        print ('Creating pir file using all templates ({n})'.format(
            n = len(selected_templates)))

    query_grid = create_query_grid(selected_templates) # load query grid
    print ('query_grid')
    print(query_grid)
    gapless_query_grid = create_gapless_grid(query_grid) # remove gaps
    print ('gapless_query_grid')
    print(gapless_query_grid)
    processed_query_grid = process_query_grid(query_grid, gapless_query_grid) # insert gaps
    ##processed_query_grid = process_query_grid(query_grid, query_grid) # insert gaps
    print ('processed_query_grid')
    print (processed_query_grid)
    glob_seq = derive_global_seq(processed_query_grid, query_name, query_chain) # derive query sequence
    template_grid = create_template_grid(selected_templates) # create template grid
    print ('template_grid')
    print (template_grid)
    processed_template_grid = process_template_grid(query_grid, template_grid) # insert gaps to template sequnces
    print ('processed_query_grid')
    print (processed_query_grid)
    print ('hzhu processed_template_grid')
    print (processed_template_grid)
    final_grid = compare_with_cifs(processed_template_grid, args.cifs, args.output, args.c, args.r) # compare with atom section of cifs
    remove_self_alignment(final_grid, query_name) # remove self alignment if any
    write_to_file([glob_seq, final_grid.display()], args.pir)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
Parser for hhr result files created with hhblits|hhsearch|hhalign -o <hhr_file>
"""


import sys
from collections import namedtuple


__author__ = 'Markus Meier (markus.meier@mpibpc.mpg.de)'
__version__ = '1.0'
__license__ = "GPL-3"


hhr_alignment = namedtuple('hhr_alignment', ['query_id', 'query_length', 'query_neff',
                                             'template_id', 'template_length', 'template_info',
                                             'template_neff', 'query_ali', 'template_ali',
                                             'start', 'end', 'probability', 'evalue', 'score',
                                             'aligned_cols', 'identity', 'similarity', 'sum_probs'])


class HHRFormatError(Exception):
    def __init__(self, value):
        self.value = "ERROR: "+value

    def __str__(self):
        return repr(self.value)


def get_sequence_name(header):
    name = header.replace(">", "").split()[0]
    return name


def parse_result(lines):
    results = []

    query_id = None
    query_length = None
    query_neff = None
    query_seq = []
    template_id = None
    template_length = None
    template_seq = []
    template_info = None
    query_start = None
    query_end = None
    template_start = None
    template_end = None
    probability = None
    evalue = None
    score = None
    identity = None
    similarity = None
    template_neff = None
    sum_probs = None
    aligned_cols = None

    skipped_ali_tags = ["ss_dssp", "ss_pred", "Consensus"]

    is_alignment_section = False

    for line in lines:
        if(line.startswith("Query")):
            query_id = line.split()[1]
        elif(line.startswith("Match_columns")):
            query_length = int(line.split()[1])
        elif(line.startswith("Neff")):
            query_neff = float(line.split()[1])
        elif(is_alignment_section and (line.startswith("No") or line.startswith("Done!"))):
            if query_start is not None:
                result = hhr_alignment(query_id, query_length, query_neff,
                                       template_id, template_length, template_info, template_neff,
                                       query_seq, template_seq, (query_start, template_start),
                                       (query_end, template_end), probability, evalue, score,
                                       aligned_cols, identity, similarity, sum_probs)
                results.append(result)
            template_id = None
            template_info = None
            query_seq = []
            template_seq = []

            query_start = None
            query_end = None
            template_start = None
            template_end = None
        elif(line.startswith("Probab")):
            tokens = line.split()
            probability = float(tokens[0].split("=")[1])
            evalue = float(tokens[1].split("=")[1])
            score = float(tokens[2].split("=")[1])
            aligned_cols = int(tokens[3].split("=")[1])
            identity = float(tokens[4].split("=")[1].replace("%", "")) / 100.0
            similarity = float(tokens[5].split("=")[1])
            sum_probs = float(tokens[6].split("=")[1])
            if(len(tokens) > 7):
                template_neff = float(tokens[7].split("=")[1])
            continue
        elif(line.startswith(">")):
            is_alignment_section = True
            template_id = line[1:].split()[0]
            template_info = line
        elif(line.startswith("Q")):
            tokens = line.split()
            if(tokens[1] in skipped_ali_tags):
                continue

            try:
                token_2 = tokens[2].replace("(", "").replace(")", "")
                token_2 = int(token_2)
            except:
                raise HHRFormatError(("Converting failure of start index ({}) "
                                      "of query alignment").format(tokens[2]))

            if query_start is None:
                query_start = token_2
            query_start = min(query_start, token_2)

            try:
                token_4 = tokens[4].replace("(", "").replace(")", "")
                token_4 = int(token_4)
            except:
                raise HHRFormatError(("Converting failure of end index ({}) "
                                      "of query alignment").format(tokens[4]))

            if query_end is None:
                query_end = token_4
            query_end = max(query_end, token_4)
            query_seq.append(tokens[3])
        elif(line.startswith("T")):
            tokens = line.split()
            if(tokens[1] in skipped_ali_tags):
                continue
            template_seq.append(tokens[3])

            try:
                token_2 = tokens[2].replace("(", "").replace(")", "")
                token_2 = int(token_2)
            except:
                raise HHRFormatError(("Converting failure of start index ({}) "
                                      "of template alignment").format(tokens[2]))

            if template_start is None:
                template_start = token_2
            template_start = min(template_start, token_2)

            try:
                token_4 = tokens[4].replace("(", "").replace(")", "")
                token_4 = int(token_4)
            except:
                raise HHRFormatError(("Converting failure of end index ({}) "
                                      "of template alignment").format(tokens[4]))

            if template_end is None:
                template_end = token_4
            template_end = max(template_end, token_4)

            try:
                token_5 = tokens[4].replace("(", "").replace(")", "")
                token_5 = int(token_5)
            except:
                raise HHRFormatError(("Converting failure of template length ({}) "
                                      "in template alignment").format(tokens[5]))
            template_length = token_5


    if(template_id is not None and query_start is not None):
        result = hhr_alignment(query_id, query_length, query_neff,
                               template_id, template_length, template_info, template_neff,
                               "".join(query_seq), "".join(template_seq), (query_start, template_start),
                               (query_end, template_end), probability, evalue, score,
                               aligned_cols, identity, similarity, sum_probs)
        results.append(result)

    return results


def read_result(input_file):
    with open(input_file) as fh:
        lines = fh.readlines()
        return parse_result(lines)


def main():
    counter = 0
    for result in read_result(sys.argv[1]):
        print("Alignment " + str(counter) + "\t evalue: " + str(result.evalue) +
              "\t probability: " + str(result.probability))

        print(result.query_id + "\t" + str(result.start[0]) + "\t" +
              result.query_ali + "\t" +
              str(result.end[0]))

        print(result.template_id + "\t" + str(result.start[1]) + "\t" +
              result.template_ali + "\t" +
              str(result.end[1]))

        counter += 1


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
Created in Jun 2016

@author: Harald Voehringer
"""


from itertools import groupby
from collections import namedtuple, defaultdict
import textwrap

DEBUG = False
FASTA = namedtuple('FASTA', ['header', 'seq'])
PDBDATA = namedtuple('PDBDATA', ['entry', 'res', 'rfr', 'comp', 'met'])

def as_pairs(fasta):
    """ Reads in fasta headers as defined """

    for header, group in groupby(fasta, lambda x: x[0] == '>'):
        if header:
            line = next(group)
            identifier = line.split(' ')[0].split('>')[1]
            header_line = str(line.strip())

        else:
            sequence = ''.join(line.strip() for line in group).upper()
            data = FASTA(header_line, sequence)

            yield identifier, data
 
def read_fasta(fasta_file):
    """ Reads in fasta sequences with help of the as_pairs function."""
    fasta_dic = dict()
    duplicate = 0
    
    with open(fasta_file) as fasta:
        for title, sequence in as_pairs(fasta):
            if title not in fasta_dic:
                fasta_dic[title] = sequence
            else:
                duplicate +=1
    
    if duplicate != 0:
        print ('! Warning found duplicate {num} fasta.'.format(
            num = duplicate))
        
    return fasta_dic

def read_fasta_annotations(fname):
    """ Reads the information about resolution, R-free and completness which 
    be outputed by cif2fasta."""

    annotations = dict()

    with open(fname) as fh:
        for line in fh:
            if len(line) > 0 and line[0] == '#':
                continue

            identifier, res, r_free, comp, method = line.strip().split('\t')

            try:
                res = float(res)
            except ValueError:
                res = None

            try:
                r_free = float(r_free)
            except ValueError:
                r_free = None

            try:
                comp = float(comp)
            except ValueError:
                comp = None

            annotations[identifier] = PDBDATA(identifier, res, r_free, comp, method)

    return annotations

def read_cluster(cluster_file):
    """ Reads in the clusters (this is the output of MMseqs). """
    cluster = defaultdict(set)

    with open(cluster_file) as fh:
        for line in fh:
            exemplar, node = line.split()
            
            if node in cluster[exemplar]:
                raise Exception('! Please check the clustering procedure: {n} was found twice in cluster {e}.'.format(
                    n = node,
                    e = exemplar))
            else:
                cluster[exemplar].add(node)

    return cluster

def read_pdblist(in_file):

    pdb_list = set()
    
    with open(in_file) as fh:
        for line in fh:
            if line.startswith('#'):
                continue

            strip_line = line.strip()
            pdb_code = strip_line.split('_')[0]

            # this is a very very basic check
            if len(pdb_code) != 4:
                print ('! Warning: {line} seems to be an incorrect identifer. Skipping it.'.format(
                    line = strip_line))
                continue
            
            pdb = strip_line.upper()
            pdb_list.add(pdb)

    return pdb_list

def select_sequences(clusters, annotations):

    selected_sequences = set()

    for idx, cluster in enumerate(clusters):
        
        nodes = [ annotations[entry] for entry in clusters[cluster] ]
        
        if DEBUG:
            print ('Processing Cluster {c} ({i}): {m}'.format(
                c = cluster,
                i = idx, 
                m = ', '.join([x.entry for x in nodes])))

        # select the best entries of the cluster
        best_res = float('inf')
        best_entry_res = None

        best_rfr = float('inf')
        best_entry_rfr = None

        best_comp = -float('inf')
        best_entry_comp = None

        # iterate through each entry in nodes while selecting the representative sequence
        for node in nodes:
            
            if (node.res is not None) and (node.res < best_res):
                best_res = node.res
                best_entry_res = node.entry

            if (node.rfr is not None) and (node.rfr < best_rfr):
                best_rfr = node.rfr
                best_entry_rfr = node.entry

            if (node.comp is not None) and (node.comp > best_comp):
                best_comp = node.comp
                best_entry_comp = node.entry

        if best_entry_res is not None:
            selected_sequences.add(best_entry_res)
            
            if DEBUG:
                print (' - Selected {n} (best resolution = {r}).'.format(
                    n = best_entry_res,
                    r = best_res))

        if best_entry_rfr is not None:
            selected_sequences.add(best_entry_rfr)
            
            if DEBUG:
                print (' - Selected {n} (best R-free = {r}).'.format(
                    n = best_entry_rfr,
                    r = best_rfr))
        
        if best_entry_comp is not None:
            selected_sequences.add(best_entry_comp)
            
            if DEBUG:
                print (' - Selected {n} (best completness = {r}).'.format(
                    n = best_entry_comp,
                    r = best_comp))    

        if best_entry_res == None and best_entry_rfr == None and best_entry_comp == None:
            print ('! Warning: Did not find any representative entry for cluster {c}.'.format(
                c = cluster))
    
    return selected_sequences

def write_sequences(out_file, fasta_db, selected_sequences):
    """ Writes selected sequences to a fasta file."""
    with open(out_file, 'w') as fh:
        for seq in selected_sequences:
            
            fasta_entry = '{h}\n{seq}\n'.format(
                h = fasta_db[seq].header,
                seq = '\n'.join(textwrap.wrap(fasta_db[seq].seq, 80)))
            
            fh.write(fasta_entry)

def arg():
    import argparse
    description = """
    pdbfilter.py selects from sequence clusters (determined by MMSeqs) the sequences 
    which have the best resolution, R-free factor and/or completness and writes them to a fasta file. 
    """.replace('\n', '')
    
    epilog = '2016 Harald Voehringer'
    # Initiate a ArgumentParser Class
    parser = argparse.ArgumentParser(description = description, epilog = epilog)
    
    # Call add_options to the parser
    parser.add_argument('fasta', help = 'input fasta file (created by cif2fasta.py)', metavar = 'FILE')    
    parser.add_argument('cluster', help = 'sequence clusters (MMseqs)', metavar = 'FILE')
    parser.add_argument('annotations', help = 'annotations file (created by cif2fasta using the -p flag, contains information about resolution, R-free and completness of sequences).', metavar = 'FILE')
    parser.add_argument('out_file', help = 'output fasta file', metavar = 'FILE')
    parser.add_argument('-i', '--include', help = 'include PDB chains', metavar = 'FILE')
    parser.add_argument('-r', '--remove', help = 'exclude PDB chains', metavar = 'FILE')

    parser.add_argument('-v', '--verbose', action = 'store_true', help = 'verbose mode')


    return parser


def main():
    import sys
    parser = arg()
    args = parser.parse_args(sys.argv[1:])

    global DEBUG
    
    if args.verbose:
        DEBUG = True

    fasta = read_fasta(args.fasta)
    clu70 = read_cluster(args.cluster)
    annot = read_fasta_annotations(args.annotations)

    print ("Found {i} clusters.".format(
        i = len(clu70.keys())))

    # choose representative sequences from clusters
    selection = select_sequences(clu70, annot)

    # make sure that pdbs specified in the argument are included
    if args.include:
        to_include = read_pdblist(args.include)
        for pdb in to_include:
            if pdb in fasta.keys():
                if pdb not in selection:
                    if DEBUG:
                        print ('Adding {p}.'.format(
                            p = pdb))
                    selection.add(pdb)
            else:
                print ('! Warning: {p} was not found in input fasta.'.format(
                    p = pdb))

    # removes entries
    if args.remove:
        to_remove = read_pdblist(args.remove)
        for pdb in to_remove:
            if pdb in selection:
                if DEBUG:
                    print ('Removing {p}.'.format(
                        p = pdb))
                selection.remove(pdb)
        
    # write them to file
    write_sequences(args.out_file, fasta, selection)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
    hhbsuite.py
    Creates HH-suite database files from A3M and HHM files 
    Usage: Usage: python hhsuite_db.py -o <db_name> [-ia3m <a3m_dir>] [-ihhm <hhm_dir>] [-ics <cs_dir>] [more_options]

    HHsuite version 3.0.0 (15-03-2015)

    Reference: 
    Remmert M., Biegert A., Hauser A., and Soding J.
    HHblits: Lightning-fast iterative protein sequence searching by HMM-HMM alignment.
    Nat. Methods, epub Dec 25, doi: 10.1038/NMETH.1818 (2011).

    (C) Johannes Soeding, 2012

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    We are very grateful for bug reports! Please contact us at soeding@mpibpc.mpg.de
"""


from optparse import OptionParser
from glob import glob
from subprocess import check_call
import ffindex
import a3m
import tempfile
import os
import sys
import shutil


def write_glob_to_file(glob_expr, filename):
  fh = open(filename, "w")
  for f in glob(glob_expr):
    fh.write(f+"\n")
  fh.close()


def write_set_to_file(set, filename):
  fh = open(filename, "w")
  for f in set:
    fh.write(f+"\n")
  fh.close()


def remove_files_from_index(files_file, database_index_path):
  if(os.path.exists(database_index_path) and os.path.getsize(database_index_path) > 0):
    check_call(" ".join(["ffindex_modify -us -f ", files_file, database_index_path]), shell=True)


def build_ffindex_database(files_file, data_path, index_path):
  check_call(" ".join(["ffindex_build","-s", data_path, index_path, "-f", files_file]), shell=True)


def optimize_database(database_data_path, database_index_path):
  if os.path.exists(database_data_path) and os.path.exists(database_index_path) and os.path.getsize(database_data_path) > 0 and os.path.getsize(database_index_path) > 0: 
    check_call(" ".join(["ffindex_build", "-as", database_data_path+".optimized", database_index_path+".optimized", "-d", database_data_path, "-i", database_index_path]), shell=True)
    shutil.move(database_data_path+".optimized", database_data_path)
    shutil.move(database_index_path+".optimized", database_index_path)


def get_large_a3ms(a3m_base_path):
  entries = ffindex.read_index(a3m_base_path+".ffindex")
  data = ffindex.read_data(a3m_base_path+".ffdata")
  
  large_alignments = set()
  for entry in entries:
    lines = ffindex.read_lines(entry, data)
    alignment = a3m.A3M_Container()
    try:
      alignment.read_a3m_from_lines(lines)
      
      if alignment.number_sequences > 50:
        large_alignments.add(entry.name)
    except:
      sys.stderr.write("Warning: A3M "+entry.name+" is corrupted!\n")

  if len(entries) > 0 and len(large_alignments) == 0:
    large_alignments.add(entries[0].name)

  return large_alignments


def calculate_hhm(threads, a3m_base_path, hhm_base_path):
  tmp_dir = tempfile.mkdtemp()
  
  large_a3ms = get_large_a3ms(a3m_base_path)
  large_a3m_index = os.path.join(tmp_dir, "large.ffindex")

  a3m_index = read_ffindex(a3m_base_path+".ffindex")
  write_subset_index(a3m_index, large_a3ms, large_a3m_index)
  
  check_call(" ".join(["mpirun", "-np", threads, "ffindex_apply_mpi", a3m_base_path+".ffdata", large_a3m_index, "-d", hhm_base_path+".ffdata", "-i", hhm_base_path+".ffindex", "--", "hhmake", "-v", str(0), "-i", "stdin", "-o" ,"stdout"]), shell=True)


def calculate_cs219(threads, a3m_db_base, cs_db_base):
  hhlib_environment = os.environ['HHLIB']
  if(hhlib_environment):
    #TODO: check
    check_call(" ".join(["cstranslate", "-A ", os.path.join(hhlib_environment, "data/cs219.lib"), "-D", os.path.join(hhlib_environment, "data/context_data.lib"), "-x 0.3 -c 4 --ffindex", "-i", a3m_db_base, "-o", cs_db_base, "-I a3m -b"]), env=dict(os.environ, OMP_NUM_THREADS=threads), shell=True)
  else:
    sys.err("ERROR: HHLIB environment variable not set! See manual!\n")
    exit(1)
    #TODO throw error


def merge_databases(source_data_path, source_index_path, dest_data_path, dest_index_path):
  check_call(["ffindex_build", "-as", "-d", source_data_path, "-i", source_index_path, dest_data_path, dest_index_path])


def sort_database(data_path, index_path):
  check_call(["ffindex_build", "-as", data_path, index_path])


def read_ffindex(file):
  fh = open(file, "r")
  
  index = []
  for line in fh:
    index.append(line.rstrip().split())
  
  fh.close()
  
  return index


def write_subset_index(index, subset, output_file):
  fh = open(output_file, "w")
  for entry in index:
    if entry[0] in subset:
      fh.write("{name:.64}\t{offset}\t{length}\n".format(name=entry[0], offset=entry[1], length=entry[2]))
  fh.close()


def is_sorted(index):
  for i in range(len(index)-1):
    if(index[i][0] > index[i+1][0]):
      return False
  return True


def get_duplicates(index):
  duplicates = set()
  entries = set()
  
  for entry in index:
    if(entry[0] in entries):
      duplicates.add(entry[0])
      
    entries.add(entry[0])

  return duplicates


def get_missing(index, cmp_index):
  missing = set()
  
  index_set = set()
  for entry in index:
    index_set.add(entry[0])
    
  cmp_index_set = set()
  for entry in cmp_index:
    cmp_index_set.add(entry[0])
  
  for key in cmp_index_set:
    if key not in index_set:
      missing.add(key)
  
  return missing


def get_overhead(index, cmp_index):
  overhead = set()
  
  index_set = set()
  for entry in index:
    index_set.add(entry[0])
    
  cmp_index_set = set()
  for entry in cmp_index:
    cmp_index_set.add(entry[0])
  
  for key in index_set:
    if key not in cmp_index_set:
      overhead.add(key)
  
  return overhead


def handle_duplicates(suffix, calculate, threads, db_basename, force_mode):
  index = read_ffindex(db_basename+"_"+suffix+".ffindex")
  duplicates = get_duplicates(index)
  
  if(suffix == "a3m" and len(duplicates) > 0):
    sys.stderr.write("ERROR: "+db_basename+"_a3m.ffindex contains duplicates!\n")
    sys.stderr.write("ERROR: Your database is broken!\n")
    exit(1);
    
  if(len(duplicates) == 0):
    return
  
  for duplicate in duplicates:
    sys.stderr.write("WARNING: "+db_basename+"_"+suffix+".ffindex contains duplicate "+duplicate+"!\n")

  
  if force_mode:
    tmp_dir = tempfile.mkdtemp()
    
    try:
      sys.stderr.write("WARNING: remove duplicates and recalculate them from their corresponding a3m's!\n")
      a3m_index = read_ffindex(db_basename+"_a3m.ffindex")
      
      duplicates_a3m_base = os.path.join(tmp_dir, "duplicates_a3m")
      duplicates_a3m_index_file = os.path.join(tmp_dir, "duplicates_a3m.ffindex")
      duplicates_a3m_data_file = os.path.join(tmp_dir, "duplicates_a3m.ffdata")
      write_subset_index(a3m_index, duplicates, duplicates_a3m_index_file)
      os.symlink(db_basename+"_a3m.ffdata", duplicates_a3m_data_file)
      
      duplicates_new_base = os.path.join(tmp_dir, "duplicates_new")
      duplicates_new_index_file = os.path.join(tmp_dir, "duplicates_new.ffindex")
      duplicates_new_data_file = os.path.join(tmp_dir, "duplicates_new.ffdata")
    
      duplicates_index_file = os.path.join(tmp_dir, "duplicates.dat")
      write_set_to_file(duplicates, duplicates_index_file)
      remove_files_from_index(duplicates_index_file, db_basename+"_"+suffix+".ffindex")
      
      calculate(threads, duplicates_a3m_base, duplicates_new_base)
      sort_database(duplicates_new_data_file, duplicates_new_index_file)
      
      merge_databases(duplicates_new_data_file, duplicates_new_index_file, 
        db_basename+"_cs219.ffdata", db_basename+"_cs219.ffindex")
      sort_database(db_basename+"_cs219.ffdata", db_basename+"_cs219.ffindex")
      optimize_database(db_basename+"_cs219.ffdata", db_basename+"_cs219.ffindex")
    finally:
      shutil.rmtree(tmp_dir)
  else:
    sys.stderr.write("You may try to use the option --force to fix the database!\n")

    
def handle_unsorted(suffix, db_basename, force_mode):
  index = read_ffindex(db_basename+"_"+suffix+".ffindex")
  if(not is_sorted(index)):
    sys.stderr.write("Index "+db_basename+"_"+suffix+".ffindex is unsorted!\n")
    if force_mode:
      sys.stderr.write("Try to sort unsorted index "+db_basename+"_"+suffix+".ffindex!\n")
      sort_database(db_basename+"_"+suffix+".ffdata", db_basename+"_"+suffix+".ffindex")
    else:
      sys.stderr.write("You may try to use the option --force to fix the database!\n")

    
def handle_missing(suffix, calculate, db_basename, threads, force_mode):
  index = read_ffindex(db_basename+"_"+suffix+".ffindex")
  a3m_index = read_ffindex(db_basename+"_a3m.ffindex")
  
  missing = get_missing(index, a3m_index)
  
  if(len(missing) == 0):
    return
  
  for f in missing:
    sys.stderr.write("WARNING: Missing entry "+f+" in "+db_basename+"_"+suffix+".ff{data,index}!\n")
  
  if force_mode:
    sys.stderr.write("WARNING: Try to calculate missing entries!\n")
    tmp_dir = tempfile.mkdtemp()
    
    try:
      missing_base = os.path.join(tmp_dir, "missing_"+suffix)
      missing_index_file = os.path.join(tmp_dir, "missing_"+suffix+".ffindex")
      missing_data_file = os.path.join(tmp_dir, "missing_"+suffix+".ffdata")
    
      missing_a3m_base = os.path.join(tmp_dir, "missing_a3m")
      missing_a3m_index_file = os.path.join(tmp_dir, "missing_a3m.ffindex")
      missing_a3m_data_file = os.path.join(tmp_dir, "missing_a3m.ffdata")
      write_subset_index(a3m_index, missing, missing_a3m_index_file)
      os.symlink(os.path.abspath(db_basename+"_a3m.ffdata"), missing_a3m_data_file)
      
      calculate(threads, missing_a3m_base, missing_base)
    
      merge_databases(missing_data_file, missing_index_file, db_basename+"_"+suffix+".ffdata", db_basename+"_"+suffix+".ffindex")
      optimize_database(db_basename+"_"+suffix+".ffdata", db_basename+"_"+suffix+".ffindex")
    finally:
      shutil.rmtree(tmp_dir)
  else:
    sys.stderr.write("You may try to use the option --force to fix the database!\n")


def handle_overhead(suffix, db_basename, force_mode):
  index = read_ffindex(db_basename+"_"+suffix+".ffindex")
  a3m_index = read_ffindex(db_basename+"_a3m.ffindex")

  overhead = get_overhead(index, a3m_index)

  #delete overhead cs219 files
  if(len(overhead) == 0):
    return

  for f in overhead:
    sys.stderr.write("WARNING: Entry "+f+" from "+db_basename+"_"+suffix+".ff{data,index} has no corresponding entry in the a3m database!\n")

  if force_mode:
    sys.stderr.write("WARNING: Try to fix overhead entries!\n")
    tmp_dir = tempfile.mkdtemp()
    
    try:
      index_file = os.path.join(tmp_dir, "to_delete.dat")
      write_set_to_file(overhead, index_file)
      remove_files_from_index(index_file, db_basename+"_"+suffix+".ffindex")
      optimize_database(db_basename+"_"+suffix+".ffdata", db_basename+"_"+suffix+".ffindex")
    finally:
      shutil.rmtree(tmp_dir)
  else:
    sys.stderr.write("You may try to use the option --force to fix the database!\n")


def check_a3m_format(db_basename, force_mode):
  entries = ffindex.read_index(db_basename+"_a3m.ffindex")
  data = ffindex.read_data(db_basename+"_a3m.ffdata")
  
  corrupted_alignments = set()
  for entry in entries:
    lines = ffindex.read_lines(entry, data)
    alignment = a3m.A3M_Container()
    try:
      alignment.read_a3m_from_lines(lines)
    except:
      corrupted_alignments.add(entry.name)
      sys.stderr.write("Warning: A3M "+entry.name+" is corrupted!\n")
  
  if len(corrupted_alignments) == 0:
    return
  
  if force_mode:
    tmp_dir = tempfile.mkdtemp()
    
    try:
      sys.stderr.write("WARNING: remove corrupted a3m's!\n")
      
      corrupted_index_file = os.path.join(tmp_dir, "corrupted.dat")
      write_set_to_file(corrupted_alignments, corrupted_index_file)
      
      for suffix in ["a3m", "cs219", "hhm"]:
        remove_files_from_index(corrupted_index_file, db_basename+"_"+suffix+".ffindex")
        sort_database(db_basename+"_"+suffix+".ffdata", db_basename+"_"+suffix+".ffindex")
        optimize_database(db_basename+"_"+suffix+".ffdata", db_basename+"_"+suffix+".ffindex")
    finally:
      shutil.rmtree(tmp_dir)
  else:
    sys.stderr.write("You may try to use the option --force to fix the database!\n")
    

def check_database(output_basename, threads, force_mode):
  if not os.path.exists(output_basename+"_a3m.ffindex") or not os.path.exists(output_basename+"_a3m.ffdata"):
    sys.stderr.write("Error: No a3m database found!")
    exit(1)
    
  if not os.path.exists(output_basename+"_cs219.ffindex") or os.path.exists(output_basename+"_cs219.ffdata"):
    open(output_basename+"_cs219.ffindex", 'a').close()
    open(output_basename+"_cs219.ffdata", 'a').close()
    
  if not os.path.exists(output_basename+"_hhm.ffindex") or os.path.exists(output_basename+"_hhm.ffdata"):
    calculate_hhm(threads, output_basename+"_a3m", output_basename+"_hhm")

  check_a3m_format(output_basename, force_mode)
  #TODO check hhm's...
  #TODO check cs219...
  
  handle_unsorted("a3m", output_basename, force_mode)
  handle_unsorted("hhm", output_basename, force_mode)
  handle_unsorted("cs219", output_basename, force_mode)
  
  handle_duplicates("a3m", calculate_hhm, threads, output_basename, force_mode)
  handle_duplicates("hhm", calculate_hhm, threads, output_basename, force_mode)
  handle_duplicates("cs219", calculate_cs219, threads, output_basename, force_mode)

  handle_missing("cs219", calculate_cs219, output_basename, threads, force_mode)
  
  handle_overhead("hhm", output_basename, force_mode)
  handle_overhead("cs219", output_basename, force_mode)


def add_new_files(globular_expression, suffix, output_basename):
  tmp_dir = tempfile.mkdtemp()

  files = set(glob(globular_expression))
  
  if(len(files) == 0):
    return

  try:
    files_index = os.path.join(tmp_dir, "files.dat")
    write_set_to_file(files, files_index)
    
    new_base = os.path.join(tmp_dir, "new")
    new_index_file = new_base + ".ffindex"
    new_data_file = new_base + ".ffdata"
    build_ffindex_database(files_index, new_data_file, new_index_file)
  
    output_index_file = output_basename+"_"+suffix+".ffindex"
    output_data_file = output_basename+"_"+suffix+".ffdata"
    remove_files_from_index(files_index, output_index_file)
    merge_databases(new_data_file, new_index_file, output_data_file, output_index_file)
    optimize_database(output_data_file, output_index_file)
    
    if(suffix == "a3m"):
      for other_suffix in ["hhm", "cs219"]:
        output_index_file = output_basename+"_"+other_suffix+".ffindex"
        output_data_file = output_basename+"_"+other_suffix+".ffdata"
        if os.path.exists(output_index_file) and os.path.exists(output_data_file) and os.path.getsize(output_data_file) > 0:
          remove_files_from_index(files_index, output_index_file)
          optimize_database(output_data_file, output_index_file)
      
  finally:
    shutil.rmtree(tmp_dir)


def opt():
  parser = OptionParser()
  parser.add_option("--ia3m", dest="a3m_files_glob",
    help="Glob for a3m files", metavar="<GLOB>")
  parser.add_option("--ics219", dest="cs_files_glob",
    help="Glob for cs219 files", metavar="<GLOB>")
  parser.add_option("--ihhm", dest="hhm_files_glob",
    help="Glob for hhm files", metavar="<GLOB>")
  parser.add_option("-o", metavar="FILE", dest="output_basename", 
    help="Output hhsuite database basename")
  parser.add_option("--cpu", metavar="INT", dest="nr_cores", 
    help="Number of threads that shall used")
  parser.add_option("--force", 
    action="store_true", dest="force_mode", default=False,
    help="Try to fix problems with the database")


  return parser
  

def check_options(options, parser):
  if not options.nr_cores:
    sys.stderr.write("Please use --cpu!\n")
    parser.print_help()
    sys.exit(1)
    
  if not options.output_basename:
    sys.stderr.write("Please use -o!\n")
    parser.print_help()
    sys.exit(1)


def main():
  parser = opt()
  (options, args) = parser.parse_args()
  check_options(options, parser)
  
  #Important to do a3m's first... deleting out of date hhm's and cs219's
  if(options.a3m_files_glob):
    add_new_files(options.a3m_files_glob, "a3m", options.output_basename)
    
  if(options.hhm_files_glob):
    add_new_files(options.hhm_files_glob, "hhm", options.output_basename)

  if(options.cs_files_glob):
    add_new_files(options.cs_files_glob, "cs219", options.output_basename)

  check_database(options.output_basename, options.nr_cores, options.force_mode)
  

if __name__ == "__main__":
  main()
  



#!/usr/bin/env python

"""
Created on Mon Jun 15 21:49:32 2015

@author: Harald Voehringer
"""

from __future__ import print_function
import sys, os, glob, gzip, textwrap, itertools
from optparse import OptionParser
from collections import defaultdict
from os.path import splitext
from pdbx.reader.PdbxReader import PdbxReader
from multiprocessing import Pool

DEBUG_MODE = False
MIN_SEQ_LEN = None
SCOP_LIBRARY = False

THREE2ONE = {
        'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K', 'ILE': 'I', 'PRO': 'P', 
        'THR': 'T', 'PHE': 'F', 'ASN': 'N', 'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 
        'TRP': 'W', 'ALA': 'A', 'VAL': 'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M', 'MSE': 'M',
        'HYP': 'P', 'MLY': 'K', 'SEP': 'S', 'TPO': 'T', 'CSO': 'C', 'PTR': 'Y', 'KCX': 'K',
        'CME': 'C', 'CSD': 'A', 'CAS': 'C', 'MLE': 'L', 'DAL': 'A', 'CGU': 'E', 'DLE': 'L',
        'FME': 'M', 'DVA': 'V', 'OCS': 'C', 'DPR': 'P', 'MVA': 'V', 'TYS': 'Y', 'M3L': 'K',
        'SMC': 'C', 'ALY': 'K', 'CSX': 'C', 'DCY': 'C', 'NLE': 'L', 'DGL': 'E', 'DSN': 'S',
        'CSS': 'C', 'DLY': 'K', 'MLZ': 'K', 'DPN': 'F', 'DAR': 'R', 'PHI': 'F', 'IAS': 'D',
        'DAS': 'D', 'HIC': 'H', 'MP8': 'P', 'DTH': 'T', 'DIL': 'I', 'MEN': 'N', 'DTY': 'Y',
        'CXM': 'M', 'DGN': 'G', 'DTR': 'W', 'SAC': 'S', 'DSG': 'N', 'MME': 'M', 'MAA': 'A',
        'YOF': 'Y', 'FP9': 'P', 'FVA': 'V', 'MLU': 'L', 'OMY': 'Y', 'FGA': 'E', 'MEA': 'F',
        'CMH': 'C', 'DHI': 'H', 'SEC': 'C', 'OMZ': 'Y', 'SCY': 'C', 'MHO': 'M', 'MED': 'M',
        'CAF': 'C', 'NIY': 'Y', 'OAS': 'S', 'SCH': 'C', 'MK8': 'L', 'SME': 'M', 'LYZ': 'K'
    }

CANONICAL_RESIDUES = set(['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P',
      'Q', 'R', 'S', 'T', 'V', 'W', 'Y'])

class CIF2FASTA(object):
    
    def __init__(self, cif_path):
        self.cif_path = cif_path
        self.block = self.open_cif()
            
    def open_cif(self):
        """ Assumes a mmCIF or gzipped mmCIF file and returns a data block used for subsequent procedures. """
        # The "usual" procedure to open a mmCIF with pdbX/mmCIF

        data = []
        try:
            with gzip.open(self.cif_path) as cif_fh:
                reader = PdbxReader(cif_fh)
                reader.read(data)
        except IOError:
            with open(self.cif_path) as cif_fh:
                reader = PdbxReader(cif_fh)
                reader.read(data)
        if len(data) == 0:
            return None
        else:
            return data[0]

    def is_valid(self):
        return self.block is not None
 
    def chain_to_seq(self):
        """Extracts the sequence of the cif from entity_poly.pdbx_seq_one_letter_code"""
        
        cif_chain_to_seq = dict()
        non_polypeptide_chains = list()         
        
        try:
            entity_poly = self.block.getObj('entity_poly')
        except AttributeError:

            if DEBUG_MODE > 0:
                print ('! {pdb} Could not extract entity_poly table.'.format(
                    pdb = self.pdb_entry()))

            return False

        try:
            total_rows = entity_poly.getRowCount()
        except AttributeError:
            print ('! {pdb} Could not extract rows from entity_poly.'.format(
                pdb = self.pdb_entry()))

            return False

        for row in range(0, total_rows):

            if entity_poly.getValue('type', row) == 'polypeptide(L)':
                seq = entity_poly.getValue('pdbx_seq_one_letter_code', row)                                                            
                parsed_seq = parse_seq(seq) # removes special amino acids and newlines
                
                try:
                    chains = entity_poly.getValue('pdbx_strand_id', row)
                except ValueError:
                    
                    if total_rows == 1:
                        print ('! {pdb} Only one polypeptide chain, but no chain identifiers, setting it to ".".'.format(
                            pdb = self.pdb_entry()))
                        
                        cif_chain_to_seq['.'] = parsed_seq
                        
                        return cif_chain_to_seq

                    print ('! {pdb} Could not extract pdbx_strand_id from entity_poly table (polypeptide).'.format(
                        pdb = self.pdb_entry()))

                    return False

                chain_list = chains.split(',')
            
                for chain in chain_list:     
                    cif_chain_to_seq[chain] = parsed_seq
            
            else:
                
                try:
                    chains = entity_poly.getValue('pdbx_strand_id', row)
                except ValueError:
                    print ('! {pdb} Could not extract pdbx_strand_id from entity_poly table (non-polypeptide).'.format(
                        pdb = self.pdb_entry()))
                    
                    return False

                non_polypeptide_chains.append(chains)

        chains = list(cif_chain_to_seq.keys())
        # remove chains that contain only unknown residues
        for chain in chains:
            # this is a very odd way to check whether a string contains only a single char
            tmp_set = set(cif_chain_to_seq[chain])
            if len(tmp_set) == 1 and 'X' in tmp_set:
                print ('! Removing {pdb}_{chain} (contains only unknown residues).'.format(
                    pdb = self.pdb_entry(),
                    chain = chain))
                
                del cif_chain_to_seq[chain]
                continue
            
            if len(cif_chain_to_seq[chain]) < MIN_SEQ_LEN:
                print ('! Removing {pdb}_{chain} (sequence length < {min_len}).'.format(
                    pdb = self.pdb_entry(),
                    chain = chain,
                    min_len = MIN_SEQ_LEN))
                
                del cif_chain_to_seq[chain]

        if len(cif_chain_to_seq) != 0:
            
            if DEBUG_MODE > 1:
                print ('- Extracted chains of {pdb} {chains}.'.format(
                    pdb = self.pdb_entry(),
                    chains = ' '.join( str(chain) + ' (' + str(len(cif_chain_to_seq[chain])) + ')' for chain in cif_chain_to_seq.keys())))
                
                if len(non_polypeptide_chains) != 0:                
                    print ('- Following chains were non polypeptide chains {chains} no polypeptide chains were found.'.format(
                        chains = ', '.join(non_polypeptide_chains)))
            
            return cif_chain_to_seq
        
        else:
            if DEBUG_MODE > 0:
                print ('! {pdb} No polypeptide chains were found.'.format(
                    pdb = self.pdb_entry()))
            
        return False

    def chain_ratios(self, chain_to_seq):
        """ Tries to extract Sequence from the atom section """

        # chain_to_seq = self.chain_to_seq()

        if chain_to_seq != False:
            chain_ratios = dict()
            
            # compute the lengths of sequences found in _entity_poly
            entity_length = { chain : float(len(seq)) for chain, seq in chain_to_seq.items() }
            entity_chains = entity_length.keys()

            # load the atomsite and set up dictionary to keep track of sequences
            atom_site = self.block.getObj('atom_site')
            atom_seq = defaultdict(str)

            current_residue = 0
            # Iterate through the atomsection of the cif file
            for atom_row in range(0, atom_site.getRowCount()):

                # NMR structures contain many confomers
                try:
                    model_num = int(atom_site.getValue('pdbx_PDB_model_num', atom_row))
                except ValueError:
                    model_num = 1

                if model_num > 1:
                    continue

                atom_chain = atom_site.getValue('label_asym_id', atom_row)

                # get the alternative chain identifier too
                try:
                    alt_chain = atom_site.getValue('auth_asym_id', atom_row)
                except ValueError:
                    alt_chain = None

                # handle cases where there are no chains but only one structure
                if atom_chain == '.' and entity_chains[0] == '.':
                    atom_chain = '.'
                
                # get the residue and the residue number 
                try:
                    res_num = int(atom_site.getValue("label_seq_id", atom_row))
                except ValueError:
                    continue

                if res_num != current_residue:
                    residue = atom_site.getValue('label_comp_id', atom_row)
                    
                    try:
                        residue = THREE2ONE[residue]
                    except KeyError:
                        residue = 'X'

                    # try to get the chain identifier from alt_chain first, if this does not work use label_asym_id
                    if alt_chain is not None:
                    	atom_seq[alt_chain] += residue

                    # sometimes we find the right chain identifier not in the alt_chain
                    if not (atom_chain in atom_seq.keys()) and atom_chain is not None:
                    	atom_seq[atom_chain] += residue
                        
                    current_residue = res_num

            for chain in entity_length.keys():
                if chain in atom_seq.keys():
                    chain_ratios[chain] = len(atom_seq[chain]) / entity_length[chain]
                else:
                    chain_ratios[chain] = 0

            return chain_ratios
        else:
            return False

    def pdb_entry(self):
        """Extracts the PDB entry information of a cif file."""
        try:
            entry = self.block.getObj('entry')
            entry_id = entry.getValue('id')
            return entry_id.replace('\n', ' ')
        
        except AttributeError:
            if DEBUG_MODE > 0:
                print ('! {pdb} Could not extract id from entry.'.format(
                    pdb = self.pdb_entry()))

    def protein_description(self):
        """Extracts the protein description annotated in struct.pdbx_descriptor of the cif file."""
        try:
            # Get struct table which contains the protein description
            struct = self.block.getObj('struct')
            # Get the pdbx description and make format it appropritaly
            protein_description = struct.getValue('pdbx_descriptor')
            protein_description = protein_description.replace('\n', ' ')
            protein_description = protein_description.replace(';', ' ') # to prevent parsing errors
            
            if len(protein_description.split(' ')) >= 5: 
                protein_description = ' '.join(protein_description.split(' ')[0:5]) # maximum of 5 words in header
            
            return protein_description.strip(',')
        
        except AttributeError:
            if DEBUG_MODE > 1:
                print ('! {pdb} Could not extract pdbx_descriptor from struct table.'.format(
                    pdb = self.pdb_entry()))
            
            return False

    def compounds(self):
        """ Extracts all compounds annotated in the HETATM section of the atom
            struct table if the compound appears at least 10 times and is not water 
            (HOH)."""    
        
        atom_site = self.block.getObj('atom_site')
        compounds = {}
        
        for row in range(0, atom_site.getRowCount()):
            if atom_site.getValue('group_PDB', row) == 'HETATM':

                label_comp_id = atom_site.getValue('label_comp_id', row)
                
                if label_comp_id not in compounds.keys():
                    compounds[label_comp_id] = 1
                else:
                    compounds[label_comp_id] += 1
    
        filtered_compounds = set()

        for compound in compounds.keys():
            if compounds[compound] >= 10 and compound != 'HOH':
                filtered_compounds.add(compound)

        if len(filtered_compounds) == 0:
            return False
        else:
            return ', '.join(filtered_compounds).replace('\n', ' ')
    
    def resolution(self):
        """Extracts the resolution of the mmCIF."""

        try:
            refine = self.block.getObj('refine')
            resolution = refine.getValue('ls_d_res_high')
            
            try: 
                resolution = float(resolution)
            except ValueError:
                return False
            
            return resolution
        
        except AttributeError:
            if DEBUG_MODE > 1:
                print ('! {pdb} Could not extract ls_d_res_high from refine table.'.format(
                    pdb = self.pdb_entry()))            

        try:
            reflns = self.block.getObj('reflns')
            # Extract the resolution of the crystal
            resolution = reflns.getValue('d_resolution_high')
            
            try:
                resolution = float(resolution)
            except ValueError:
                return False
            
            return resolution
        
        except AttributeError:
            if DEBUG_MODE > 1:
                print ('! {pdb} Could not extract d_resolution_high from reflns table.'.format(
                    pdb = self.pdb_entry()))

        # This is true for some Electron Microscopy structures
        try:
            em_3d = self.block.getObj('em_3d_reconstruction')
            resolution = em_3d.getValue('resolution')
            
            try:
                resolution = float(resolution)
            except ValueError:
                return False
            
            return resolution
        except AttributeError:
            if DEBUG_MODE > 1:
                print ('! {pdb} Could not extract resolution from em_3d_reconstruction table.'.format(
                    pdb = self.pdb_entry())) 

        return False

    def experimental_method(self):
        """Extracts the experimental method of the mmCIF."""

        try:
            reflns = self.block.getObj('exptl')
            method = reflns.getValue('method')
            return method.replace('\n', ' ')
        
        except AttributeError:
        
            if DEBUG_MODE > 1:
                print ('! Could not extract text from exptl table.'.format(
                    pdb = self.pdb_entry()))
            
        return False

    def keywords(self):
        """Extracts the keywords of the mmCIF."""
        try:
            reflns = self.block.getObj('struct_keywords')
            keywords = reflns.getValue('text')
            # perform some string modifications
            keywords = keywords.replace('\n', ' ')
            keywords = keywords.replace(';', ' ')

            if len(keywords.split(' ')) >= 5:
                keywords = ' '.join(keywords.split(' ')[0:5]) 

            return keywords.rstrip(',')
        
        except AttributeError:
        
            if DEBUG_MODE > 1:
                print ('! {pdb} Could not extract text from struct_keywords table.'.format(
                    pdb = self.pdb_entry()))
            
        return False   

    def organism(self):
        """Extracts the organism of the mmCIF."""
        try:
            entity_src_nat = self.block.getObj('entity_src_nat')
            organsim_scientific = entity_src_nat.getValue('pdbx_organism_scientific')
            return organsim_scientific.replace('\n', ' ')
        
        except AttributeError:
        
            if DEBUG_MODE > 1:
                print ("! {pdb} Could not extract from pdbx_organism_scientific from entity_src_gen table (".format(
                    pdb = self.pdb_entry()))
            pass
            
        try:
            entity_src_gen = self.block.getObj("entity_src_gen")
            src_scientific = entity_src_gen.getValue("pdbx_gene_src_scientific_name")
            return src_scientific.replace("\n", " ")
        
        except AttributeError:
        
            if DEBUG_MODE > 1:
                print ('! {pdb} Could not extract from pdbx_gene_src_scientific_name from entity_src_gen table'.format(
                    pdb = self.pdb_entry()))
            
        return False
        
    def r_free(self):

        try:
            refine = self.block.getObj('refine')
            r_free = refine.getValue('ls_R_factor_R_free')

            try: 
                return float(r_free)
            except ValueError:
                return False

        except AttributeError:
            if DEBUG_MODE > 2:
                print ('! Could not extract R free factor ({pdb})'.format(
                    pdb = self.pdb_entry()))

        except ValueError:
            if DEBUG_MODE > 2:
                print ('! R free factor is not annotated ({pdb})'.format(
                    pdb = self.pdb_entry()))

            
        return False

# Helper functions

def parse_seq(orginal_seq):
    """Parses the cif fasta sequence and replaces non-canonical residues with their canonical counterparts."""
    seq = orginal_seq
    
    while seq.find('(') != -1: 
        start_pos = seq.find('(')
        stop_pos = seq.find(')')
        residue = seq[start_pos + 1:stop_pos]

        try:
            canonical = THREE2ONE[residue]
        except KeyError:
            canonical = 'X'

        if DEBUG_MODE > 1:
            print ('! Replaced non canonical residue {nc} with {c}'.format(
                nc = residue,
                c = canonical))
                
        if start_pos == 0:
            seq = canonical + seq[stop_pos+1:]
        elif stop_pos == len(seq):
            seq = seq[0:start_pos] + canonical 
        else:
            pre_seq = seq[0:start_pos]
            post_seq = seq[stop_pos+1:]
            seq = pre_seq + canonical + post_seq
    
    seq = seq.replace('\n', '')
    seq_array = []
    
    for c in seq:
        if c in CANONICAL_RESIDUES:
            seq_array.append(c)
        else:
            seq_array.append('X')

    seq = ''.join(seq_array)

    return seq
    
    
def get_paths(in_folder, out_folder):
    """Searches a directory and all its subdirectories for files ending 
    with a specific suffix."""

    paths = list()
    
    for root, dirs, files in os.walk(in_folder):
        for fname in files:
            if fname.endswith(".cif") or fname.endswith(".cif.gz"):
                in_path = os.path.join(root, fname)
                out_file = in_path.split('/')[-1].split('.')[0] + ".fasta"
                out_path = os.path.join(out_folder, out_file)
                #absolute_path = os.path.join(os.getcwd(), fpath)
                paths.append((in_path, out_path))
    
    return paths
        
def construct_header(cif2fasta):
    """Constructs a fasta header."""
    
    protein_description = cif2fasta.protein_description()
    keywords = cif2fasta.keywords()
    compounds = cif2fasta.compounds()
    resolution = cif2fasta.resolution()
    method = cif2fasta.experimental_method()
    organism = cif2fasta.organism()
    r_free = cif2fasta.r_free()

    # construct the header with the data given
    header_information = ''

    if protein_description:
        protein_description = protein_description.replace(';', ' ') # to prevent parsing errors
        protein_description = ' '.join(protein_description.split(' ')[0:5]) # maximum of 5 words in header
        header_information += 'DSC: ' + protein_description + '; '
    else:
        header_information += 'DSC: N/A; '
 
    # if keywords:
    #     header_information += keywords + '; '

    if method:
        header_information += 'MET: ' + method + '; '
    else:
        header_information += 'MET: N/A; '
    
    if resolution:
        header_information += 'RES: ' + resolution + '; '
    else:
        header_information += 'RES: N/A; '

    if r_free:
        header_information += 'RFR: ' + r_free + '; ' 
    else:
        header_information += 'RFR: N/A; ' 

    if organism:
        header_information += 'ORG: ' + organism + '; ' 
    else:
        header_information += 'ORG: N/A; ' 

    if compounds:
        header_information += 'HET: ' + compounds + '; '
    else:
        header_information += 'HET: N/A; '

    if SCOP_LIBRARY:
        
        try: 
            scop_idx = SCOP_LIBRARY[cif2fasta.pdb_entry()]
            if len(scop_idx) != 0:
                domains = ', '.join(scop_idx)
            else:
                domains = 'N/A'
        except KeyError:
            domains = 'N/A'
        
        header_information += 'SCOP: ' + domains + '; '

    return header_information.strip()

def create_fasta_entry(cif2fasta):
    """Combines information given in pdb_entry and strand_seq and yields a clean
    fasta file.
    """
    
    pdb_entry = cif2fasta.pdb_entry()
    header = construct_header(cif2fasta)
    chain_to_seq = cif2fasta.chain_to_seq()
    chain_ratios = cif2fasta.chain_ratios(chain_to_seq)
    #import pdb; pdb.set_trace()
    
    fasta_entry = ""
    
    if chain_to_seq and pdb_entry:
        for chain in sorted(chain_to_seq.keys()):
            if len(chain_to_seq[chain]) != 0:
                
                fasta_entry += '>{pdb}_{chain} {header} CMP: {rat:.2f}\n{seq}\n'.format(
                    pdb = pdb_entry,
                    chain = chain,
                    header = header,
                    rat = chain_ratios[chain],
                    seq = '\n'.join(textwrap.wrap(chain_to_seq[chain], 80)))
                
                # fasta_entry += '>' + pdb_entry + '_' + chain + ' ' + header + '\n' + '\n'.join(textwrap.wrap(chain_to_seq[chain], 80)) + '\n'
        return fasta_entry
    else:
        return None


def create_fasta_entry2(cif2fasta):
    """" Creates a fasta entry."""

    # Get all the information we need
    pdb_entry = cif2fasta.pdb_entry()
    protein_description = cif2fasta.protein_description()
    keywords = cif2fasta.keywords()
    compounds = cif2fasta.compounds()
    resolution = cif2fasta.resolution()
    method = cif2fasta.experimental_method()
    organism = cif2fasta.organism()
    r_free = cif2fasta.r_free()

    # get chains and sequences
    chain_to_seq = cif2fasta.chain_to_seq()
    chain_ratios = cif2fasta.chain_ratios(chain_to_seq)

    is_NMR = False
    if 'NMR' in method:
        is_NMR = True

    fasta_entry = ''
    pdb_filter = '' # If needed create entries that are parseable by pdb filter   
    
    if chain_to_seq and pdb_entry:
        for chain, seq in chain_to_seq.items():

            # check if we can find SCOP domains for the current chain
            domains = False
            
            if SCOP_LIBRARY:
                try: 
                    scop_idx = SCOP_LIBRARY[pdb_entry + '_' + chain]
                    if len(scop_idx) != 0:
                        domains = ', '.join(scop_idx)
                    else:
                        domains = False
                except KeyError:
                    domains = False

            if is_NMR:
                fasta_entry += '>{p}_{c}{d}{k}{h}{r}{{{o}}}{s}\n{seq}\n'.format(
                    p = pdb_entry,
                    c = chain,
                    d = ' ' + protein_description + ';' if protein_description else '',
                    k = ' ' + keywords + ';' if keywords else '',
                    h = ' HET: ' + compounds + ';' if compounds else '',
                    r = ' NMR ',
                    o = organism if organism else 'N/A',
                    s = ' SCOP: ' + domains if domains else '',
                    seq = '\n'.join(textwrap.wrap(seq, 80)))
                pdb_filter += '{p}_{c}\t{r}\t{f}\t{comp:.3f}\t{m}\n'.format(
                    p = pdb_entry,
                    c = chain,
                    m = method,
                    r = 'N/A',
                    f = round(r_free, 3) if r_free else 'N/A',
                    comp = round(chain_ratios[chain], 3))
            else:
                fasta_entry += '>{p}_{c}{d}{k}{h}{r}{{{o}}}{s}\n{seq}\n'.format(
                    p = pdb_entry,
                    c = chain,
                    d = ' ' + protein_description + ';' if protein_description else '',
                    k = ' ' + keywords + ';' if keywords else '',
                    h = ' HET: ' + compounds + ';' if compounds else '',
                    r = ' ' + str(resolution) + 'A ' if resolution else '',
                    o = organism if organism else 'N/A',
                    s = ' SCOP: ' + domains if domains else '',
                    seq = '\n'.join(textwrap.wrap(seq, 80)))
            # <pdb_entry>_<chain>\t<resolution>\t<r_free>\t<completness>
                pdb_filter += '{p}_{c}\t{r}\t{f}\t{comp:.3f}\t{m}\n'.format(
                    p = pdb_entry,
                    c = chain,
                    m = method,
                    r = round(resolution, 3) if resolution else 'N/A',
                    f = round(r_free, 3) if r_free else 'N/A',
                    comp = round(chain_ratios[chain], 3))

    return (fasta_entry, pdb_filter)

def parse_scop(scop_file):
    scop = defaultdict(set)

    with open(scop_file) as fh:
        for line in fh:
            if line.startswith('#'):
                continue
            
            scop_id, pdb_code, chain, scop_num = line.split('\t')[0:4]
            chain = chain.split(':')[0]
            entry = pdb_code.upper() + '_' + chain
            
            scop[entry].add(scop_num)

    return scop

def wrapper_function(paths):
    
    in_file = paths[0]
    out_file = paths[1]
    
    cif2fasta = CIF2FASTA(in_file)
    if cif2fasta.is_valid():
        fasta_entry = create_fasta_entry2(cif2fasta)
    else:
        print("Warning: Could not read %".format(in_file), file=sys.stderr)
        fasta_entry = None

    return fasta_entry

def write_to_file(line_list, fname, pdb_filter):
    """
    Input: A list containing all lines that have to be saved to the file (line_list).
    A filename (str) is the name of the file that is created.
    Output: A file containing the lines in line_list.
    """
    if pdb_filter:
        fasta_file = open(fname, 'w')
        pdb_filter = open(pdb_filter, 'w')

        pdb_filter.write('#pdb_chain\tresolution\tr_free\tcompleteness\tmethod\n')

        for line in line_list:
            if line is not None:
                fasta_file.write(line[0])
                pdb_filter.write(line[1])

        fasta_file.close()
        pdb_filter.close()
    else:
        fasta_file = open(fname, 'w')

        for line in line_list:
            if line is not None:
                fasta_file.write(line[0])

def opt():
    # Initiate a OptionParser Class
    usage = "usage: cif2fasta.py -i cif_folder -o *.fasta -c num_cores -v"
    description = "cif2fasta.py takes a folder that contains cif files as input and outputs their sequences into fasta file."
    parser = OptionParser(usage = usage, description = description)
    # Call add_options to the parser
    parser.add_option("-i", help = "input cif folder.", dest = "input_files", metavar = "DIR")
    parser.add_option("-o", help = "output fasta file.", dest = "output_files", metavar = "FILE")  
    parser.add_option("-p", help = "output PDB filter file (optional).", dest= "pdb_filter", default = False, metavar = "FILE")   
    parser.add_option("-s", help = "SCOP annotation.", dest = "scop", default = False, metavar = "FILE")   
    parser.add_option('-c', help = "number of cores (default = 1).", dest = "cores", type = int, default = 1, metavar = "INT")
    parser.add_option('-l', help = "Remove chains with a length < X (default = 30).", dest = "seq_len", type = int, default = 30, metavar = "INT")
    parser.add_option('-v', help = 'Verbose Mode (quiet = 0, full verbosity = 2).', dest = 'bool', default = 0, type = int, metavar = "INT")

    return parser
 

def main():
    parser = opt()
    (options, argv) = parser.parse_args()    
    
    global DEBUG_MODE
    
    if options.bool:
        DEBUG_MODE = options.bool

    global SCOP_LIBRARY
    
    if options.scop:
        SCOP_LIBRARY = parse_scop(options.scop)

    global MIN_SEQ_LEN

    MIN_SEQ_LEN = options.seq_len

    paths = get_paths(options.input_files, options.output_files)
    print ("Found: " + str(len(paths)) + " files.")

    if options.cores > 1:
        pool = Pool(options.cores)
        fastas = pool.map(wrapper_function, paths)
    else:
        fastas = map(wrapper_function, paths)

    write_to_file(fastas, options.output_files, options.pdb_filter)

if __name__ == "__main__":
    main()




#!/usr/bin/env python

from a3m import A3M_Container
from a3m import A3MFormatError
import sys


def main():
  filename = sys.argv[1]

  a3m = A3M_Container()
  
  if(filename.lower() == "stdin"):
    fh = sys.stdin
  else:
    fh = open(filename, "r")
    
  try:
    a3m.read_a3m(fh)
    size = a3m.get_number_sequences()
    print(size)
  except A3MFormatError as e:
    sys.stderr.write(e)
    exit(1)


if __name__ == "__main__":
  main()




#!/usr/bin/env python


class A3MFormatError(Exception):
    def __init__(self, value):
        self.value = "ERROR: "+value

    def __str__(self):
        return repr(self.value)


class A3M_Container:
    RESIDUES = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    VALID_MATCH_STATES = set(RESIDUES)
    VALID_INSERTION_STATES = set(RESIDUES.lower())
    VALID_GAP_STATES = set("-.")
    VALID_SS_CONF_STATES = set("0123456789")
    VALID_SS_STATES = set("ECH")
    VALID_DSSP_STATES = set("CHBEGITS-")

    def __init__(self):
        self.header = None
        self.annotations = dict()
        self.consensus = None
        self.sequences = []
        self.nr_match_states = None

    @property
    def number_sequences(self):
        """get the current number of protein sequences"""
        return len(self.sequences)

    def check_and_add_sequence(self, header, sequence):
        try:
            if (not self.check_and_add_annotation(header, sequence) and
                    not self.check_and_add_consensus(header, sequence)):
                self.check_sequence(sequence)
                self.sequences.append((header, sequence))
        except A3MFormatError as e:
            raise e

    def check_and_add_consensus(self, header, sequence):
        header_name = header[1:].split()[0]
        if header_name.endswith("_consensus"):
            if self.consensus:
                raise A3MFormatError("Multiple definitions of consensus!")
            else:
                self.check_sequence(sequence)
                self.consensus = (header, sequence)
                return True
        else:
            return False

    def check_and_add_annotation(self, header, sequence):
        annotation_classes = [
            ("ss_conf", self.check_ss_conf),
            ("ss_pred", self.check_ss_pred),
            ("ss_dssp", self.check_dssp)
        ]

        for (annotation_name, check) in annotation_classes:
            if(header[1:].startswith(annotation_name)):
                if(annotation_name in self.annotations):
                    raise A3MFormatError(
                        "Multiple definitions of {}!".format(annotation_name)
                    )
                elif check(sequence):
                    self.annotations[annotation_name] = sequence
                    return True
        return False

    def check_match_states(self, match_states):
        if not self.nr_match_states:
            self.nr_match_states = match_states

        if match_states == 0:
            raise A3MFormatError("Sequence with zero match states!")
        elif match_states != self.nr_match_states:
            raise A3MFormatError(
                ("Sequence with diverging number "
                 "of match states ({} vs. {})!").format(
                    match_states,
                    self.nr_match_states
                )
            )


    def check_ss_conf(self, sequence):
        count_match_states = sum((c in self.VALID_SS_CONF_STATES
                                 or c in self.VALID_GAP_STATES)
                                 for c in sequence)
        self.check_match_states(count_match_states)

        invalid_states = set(sequence) - self.VALID_SS_CONF_STATES
        invalid_states -= self.VALID_GAP_STATES

        if len(invalid_states):
            raise A3MFormatError(
                ("Undefined character(s) '{}' in predicted "
                 "secondary structure confidence!").format(invalid_states))
        else:
            return True

    def check_ss_pred(self, sequence):
        count_match_states = sum((c in self.VALID_SS_STATES
                                 or c in self.VALID_GAP_STATES)
                                 for c in sequence)
        self.check_match_states(count_match_states)

        invalid_states = set(sequence) - self.VALID_SS_STATES
        invalid_states -= self.VALID_GAP_STATES

        if len(invalid_states):
            raise A3MFormatError(
               ("Undefined character(s) '{}' in predicted "
                "secondary structure!").format(invalid_states))
        else:
            return True

    def check_dssp(self, sequence):
        count_match_states = sum(
                (c in self.VALID_DSSP_STATES) for c in sequence)
        self.check_match_states(count_match_states)

        invalid_states = set(sequence) - self.VALID_DSSP_STATES

        if len(invalid_states):
            raise A3MFormatError(
                ("Undefined character(s) '{}' in "
                 "dssp annotation!").format(invalid_states))
        else:
            return True

    def check_sequence(self, sequence):
        count_match_states = sum((c in self.VALID_MATCH_STATES
                                 or c in self.VALID_GAP_STATES)
                                 for c in sequence)
        self.check_match_states(count_match_states)


        invalid_states = set(sequence) - self.VALID_MATCH_STATES
        invalid_states -= self.VALID_GAP_STATES
        invalid_states -= self.VALID_INSERTION_STATES

        if len(invalid_states):
            raise A3MFormatError(
               ("Undefined character(s) '{}' in "
                "protein sequence!").format(invalid_states))
        else:
            return True

    def get_sub_sequence(self, sequence, limits):
        sub_sequence = []

        for (start, end) in limits:
            start_pos = 0
            pos = -1
            for i in range(len(sequence)):
                if (sequence[i] in self.VALID_MATCH_STATES or
                        sequence[i] in self.VALID_GAP_STATES):
                    pos += 1

                    if pos + 1 == start:
                        start_pos = i
                        break

            end_pos = 0
            pos = -1
            for i in range(len(sequence)):
                if (sequence[i] in self.VALID_MATCH_STATES or
                        sequence[i] in self.VALID_GAP_STATES):
                    pos += 1
                    if pos + 1 == end:
                        end_pos = i
                        break
            sub_sequence.append(sequence[start_pos:end_pos+1])

        return "".join(sub_sequence)

    def __str__(self):
        content = []

        if self.header:
            content.append(self.header)

        if self.consensus:
            content.append(self.consensus[0])
            content.append(self.consensus[1])

        for (header, sequence) in self.sequences:
            content.append(header)
            content.append(sequence)

        return "\n".join(content)

    def split_a3m(self, limits):
        new_a3m = A3M_Container()

        if self.consensus:
            new_consensus_sequence = self.get_sub_sequence(self.consensus[1],
                                                           limits)
            new_a3m.consensus = (self.consensus[0], new_consensus_sequence)

        for (header, sequence) in self.sequences:
            new_sequence = self.get_sub_sequence(sequence, limits)
            new_a3m.sequences.append((header, new_sequence))

        return new_a3m

    def read_a3m(self, fh):
        lines = fh.readlines()
        self.read_a3m_from_lines(lines)
        fh.close()

    def read_a3m_from_lines(self, lines):
        sequence_header = None
        sequence = []

        is_first_line = True

        for line in lines:
            line = line.strip()
            if len(line) == 0:
                continue
            elif line[0] == "#":
                if is_first_line:
                    self.header = line
            elif line[0] == ">":
                if sequence_header:
                    self.check_and_add_sequence(sequence_header,
                                                "".join(sequence))
                    sequence = []
                sequence_header = line.rstrip()
            else:
                sequence.append(line.strip().strip("\x00"))

            is_first_line = False

        if sequence_header:
            self.check_and_add_sequence(sequence_header, "".join(sequence))
